from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

import clearskies.configs
import clearskies.exceptions
import clearskies.parameters_to_properties
from clearskies import authentication, autodoc, typing
from clearskies.authentication import Authentication, Authorization, Public
from clearskies.endpoint import Endpoint
from clearskies.endpoint_group import EndpointGroup
from clearskies.endpoints.create import Create
from clearskies.endpoints.delete import Delete
from clearskies.endpoints.get import Get
from clearskies.endpoints.simple_search import SimpleSearch
from clearskies.endpoints.update import Update
from clearskies.functional import string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import SecurityHeader
    from clearskies.model import Column, Model, Schema


class RestfulApi(EndpointGroup):
    """
    Full CRUD operations for a model.

    This endpoint group sets up all five standard endpoints to manage a model:

     1. Create
     2. Update
     3. Delete
     4. Get
     5. List

    As such, you can set any option for all of the above endpoints.  All five endpoints are enabled by default
    but can be turned off individually.  It's important to understand that the actual API behavior is controlled
    by other endoints.  This endpoint group creates them and routes requests to them.  So, to fully understand
    the behavior of the subsequent Restful API, you have to consult the documentation for the endpoints themselves.

    For routing purposes, the create and list endpoints are reachable via the URL specified for this endpoint group
    and are separated by request method (POST for create by default, GET for list).  The update, delete, and get
    endoints all expect the id to be appended to the base URL, and then are separated by request method
    (PATCH for update, DELETE for delete, and GET for get).  See the example app and calls below:

    ```python
    import clearskies
    from clearskies.validators import Required, Unique
    from clearskies import columns


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = columns.Uuid()
        name = columns.String(validators=[Required()])
        username = columns.String(
            validators=[
                Required(),
                Unique(),
            ]
        )
        age = columns.Integer(validators=[Required()])
        created_at = columns.Created()
        updated_at = columns.Updated()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.RestfulApi(
            url="users",
            model_class=User,
            readable_column_names=["id", "name", "username", "age", "created_at", "updated_at"],
            writeable_column_names=["name", "username", "age"],
            sortable_column_names=["id", "name", "username", "age", "created_at", "updated_at"],
            searchable_column_names=["id", "name", "username", "age", "created_at", "updated_at"],
            default_sort_column_name="name",
        )
    )
    wsgi()
    ```

    Which spins up a fully functional API.  In the below usage examples we create two users, fetch
    one of them, update a user, delete the other, and then list all users.

    ```bash
    $ curl 'http://localhost:8080/users' -d '{"name":"Bob", "username": "bob", "age": 25}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "8bd9c03f-bb0c-41bd-afbc-f9526ded88f4",
            "name": "Bob",
            "username": "bob",
            "age": 25,
            "created_at": "2025-06-10T12:39:35+00:00",
            "updated_at": "2025-06-10T12:39:35+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/users' -d '{"name":"Alice", "username": "alice", "age": 22}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "16d483c6-0eb1-4104-b07b-32f3d736223f",
            "name": "Alice",
            "username": "alice",
            "age": 22,
            "created_at": "2025-06-10T12:42:59+00:00",
            "updated_at": "2025-06-10T12:42:59+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/users/8bd9c03f-bb0c-41bd-afbc-f9526ded88f4' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "8bd9c03f-bb0c-41bd-afbc-f9526ded88f4",
            "name": "Bob",
            "username": "bob",
            "age": 25,
            "created_at": "2025-06-10T12:39:35+00:00",
            "updated_at": "2025-06-10T12:39:35+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/users/16d483c6-0eb1-4104-b07b-32f3d736223f' -d '{"name":"Alice Smith", "age": 23}' -X PATCH | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "16d483c6-0eb1-4104-b07b-32f3d736223f",
            "name": "Alice Smith",
            "username": "alice",
            "age": 23,
            "created_at": "2025-06-10T12:42:59+00:00",
            "updated_at": "2025-06-10T12:45:01+00:00"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/users/8bd9c03f-bb0c-41bd-afbc-f9526ded88f4' -X DELETE | jq
    {
        "status": "success",
        "error": "",
        "data": {},
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/users/' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
            "id": "16d483c6-0eb1-4104-b07b-32f3d736223f",
            "name": "Alice Smith",
            "username": "alice",
            "age": 23,
            "created_at": "2025-06-10T12:42:59+00:00",
            "updated_at": "2025-06-10T12:45:01+00:00"
            }
        ],
        "pagination": {
            "number_results": 1,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }
    ```
    """

    """
    The endpoint class to use for managing the create operation.

    This defaults to `clearskies.endpoints.Create`.  To disable the create operation all together,
    set this to None.
    """
    create_endpoint = clearskies.configs.Endpoint(default=Create)

    """
    The endpoint class to use for managing the delete operation.

    This defaults to `clearskies.endpoints.Delete`.  To disable the delete operation all together,
    set this to None.
    """
    delete_endpoint = clearskies.configs.Endpoint(default=Delete)

    """
    The endpoint class to use for managing the update operation.

    This defaults to `clearskies.endpoints.Update`.  To disable the update operation all together,
    set this to None.
    """
    update_endpoint = clearskies.configs.Endpoint(default=Update)

    """
    The endpoint class to use to fetch individual records.

    This defaults to `clearskies.endpoints.Get`.  To disable the get operation all together,
    set this to None.
    """
    get_endpoint = clearskies.configs.Endpoint(default=Get)

    """
    The endpoint class to use to list records.

    This defaults to `clearskies.endpoints.SimpleSearch`.  To disable the list operation all together,
    set this to None.
    """
    list_endpoint = clearskies.configs.Endpoint(default=SimpleSearch)

    """
    The request method(s) to use to route to the create operation.  Default is ["POST"].
    """
    create_request_methods = clearskies.configs.SelectList(
        allowed_values=["GET", "POST", "PUT", "DELETE", "PATCH"], default=["POST"]
    )

    """
    The request method(s) to use to route to the update operation.  Default is ["PATCH"].
    """
    update_request_methods = clearskies.configs.SelectList(
        allowed_values=["GET", "POST", "PUT", "DELETE", "PATCH"], default=["PATCH"]
    )

    """
    The request method(s) to use to route to the delete operation.  Default is ["DELETE"].
    """
    delete_request_methods = clearskies.configs.SelectList(
        allowed_values=["GET", "POST", "PUT", "DELETE", "PATCH"], default=["DELETE"]
    )

    """
    The request method(s) to use to route to the get operation.  Default is ["GET"].
    """
    get_request_methods = clearskies.configs.SelectList(
        allowed_values=["GET", "POST", "PUT", "DELETE", "PATCH"], default=["GET"]
    )

    """
    The request method(s) to use to route to the create operation.  Default is ["GET"].
    """
    list_request_methods = clearskies.configs.SelectList(
        allowed_values=["GET", "POST", "PUT", "DELETE", "PATCH", "QUERY"], default=["GET", "POST", "QUERY"]
    )

    """
    The request method(s) to use to route to the create operation.  Default is ["POST"].
    """
    id_column_name = clearskies.configs.ModelColumn("model_class", default=None)

    """
    The base URL to be used for all the endpoints.
    """
    url = clearskies.configs.String(default="")

    authentication = clearskies.configs.Authentication(default=Public())
    authorization = clearskies.configs.Authorization(default=Authorization())
    output_map = clearskies.configs.Callable(default=None)
    output_schema = clearskies.configs.Schema(default=None)
    model_class = clearskies.configs.ModelClass(default=None)
    readable_column_names = clearskies.configs.ReadableModelColumns("model_class", default=[])
    writeable_column_names = clearskies.configs.WriteableModelColumns("model_class", default=[])
    searchable_column_names = clearskies.configs.SearchableModelColumns("model_class", default=[])
    sortable_column_names = clearskies.configs.ReadableModelColumns("model_class", default=[])
    default_sort_column_name = clearskies.configs.ModelColumn("model_class", required=True)
    default_sort_direction = clearskies.configs.Select(["ASC", "DESC"], default="ASC")
    default_limit = clearskies.configs.Integer(default=50)
    maximum_limit = clearskies.configs.Integer(default=200)
    group_by_column_name = clearskies.configs.ModelColumn("model_class")
    input_validation_callable = clearskies.configs.Callable(default=None)
    include_routing_data_in_request_data = clearskies.configs.Boolean(default=False)
    column_overrides = clearskies.configs.Columns(default={})
    internal_casing = clearskies.configs.Select(["snake_case", "camelCase", "TitleCase"], default="snake_case")
    external_casing = clearskies.configs.Select(["snake_case", "camelCase", "TitleCase"], default="snake_case")
    security_headers = clearskies.configs.SecurityHeaders(default=[])
    description = clearskies.configs.String(default="")
    where = clearskies.configs.Conditions(default=[])
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        writeable_column_names: list[str],
        readable_column_names: list[str],
        searchable_column_names: list[str],
        sortable_column_names: list[str],
        default_sort_column_name: str,
        read_only: bool = False,
        create_endpoint: Endpoint | None = Create,
        delete_endpoint: Endpoint | None = Delete,
        update_endpoint: Endpoint | None = Update,
        get_endpoint: Endpoint | None = Get,
        list_endpoint: Endpoint | None = SimpleSearch,
        create_request_methods: list[str] = ["POST"],
        update_request_methods: list[str] = ["PATCH"],
        delete_request_methods: list[str] = ["DELETE"],
        get_request_methods: list[str] = ["GET"],
        list_request_methods: list[str] = ["GET"],
        id_column_name: str = "",
        group_by_column_name: str = "",
        input_validation_callable: Callable | None = None,
        include_routing_data_in_request_data: bool = False,
        url: str = "",
        default_sort_direction: str = "ASC",
        default_limit: int = 50,
        maximum_limit: int = 200,
        request_methods: list[str] = ["POST"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        output_schema: Schema | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        authentication: Authentication = Public(),
        authorization: Authorization = Authorization(),
    ):
        self.finalize_and_validate_configuration()

        id_column_name = id_column_name if id_column_name else model_class.id_column_name

        # figure out which endpoints we actually need
        endpoints_to_build = []
        if not read_only:
            if create_endpoint:
                endpoints_to_build.append(
                    {
                        "class": create_endpoint,
                        "request_methods": create_request_methods,
                    }
                )
            if delete_endpoint:
                endpoints_to_build.append(
                    {
                        "class": delete_endpoint,
                        "request_methods": delete_request_methods,
                        "url_suffix": f"/:{id_column_name}",
                    }
                )
            if update_endpoint:
                endpoints_to_build.append(
                    {
                        "class": update_endpoint,
                        "request_methods": update_request_methods,
                        "url_suffix": f"/:{id_column_name}",
                    }
                )
        if get_endpoint:
            endpoints_to_build.append(
                {
                    "class": get_endpoint,
                    "request_methods": get_request_methods,
                    "url_suffix": f"/:{id_column_name}",
                }
            )
        if list_endpoint:
            endpoints_to_build.append(
                {
                    "class": list_endpoint,
                    "request_methods": list_request_methods,
                }
            )

        # and now build them!  Pass along our own kwargs to the endoints when we build them.  Now, technically, I
        # know what the kwargs are for each endpoint.  However, it would be a lot of duplication to manually
        # instantiate each endpoint and pass along all the kwargs.  So, fetch the list of kwargs from our own
        # __init__ and then compare that with the kwargs of the __init__ for each endpoint and map everything
        # automatically like that.  Then add in the individual config from above.

        # these lines take all of the arguments we were initialized with and dumps it into a dict.  It's the
        # equivalent of combining both *args and **kwargs without using either
        my_args = inspect.getfullargspec(self.__class__)
        local_variables = inspect.currentframe().f_locals  # type: ignore
        available_args = {arg: local_variables[arg] for arg in my_args.args[1:]}

        # we handle this one manually
        del available_args["url"]

        # now loop through the list of endpoints
        endpoints = []
        for endpoint_to_build in endpoints_to_build:
            # grab our class and any pre-defined configs
            endpoint_class = endpoint_to_build["class"]
            url_suffix = endpoint_to_build.get("url_suffix")

            # now get the allowed args out of the init and fill them out with our own.
            endpoint_args = inspect.getfullargspec(endpoint_class)
            nendpoint_args = len(endpoint_args.args)
            nendpoint_kwargs = len(endpoint_args.defaults) if endpoint_args.defaults else 0
            final_args: list[str] = []
            final_kwargs: dict[str, Any] = {}
            for arg in endpoint_args.args[1:]:
                if not available_args.get(arg):
                    continue
                final_kwargs[arg] = available_args[arg]

            if url_suffix:
                final_kwargs["url"] = url_suffix
            final_kwargs["request_methods"] = endpoint_to_build["request_methods"]
            endpoints.append(endpoint_class(*final_args, **final_kwargs))  # type: ignore

        super().__init__(
            endpoints,
            url=url,
            response_headers=response_headers,
            security_headers=security_headers,
            authentication=authentication,
            authorization=authorization,
        )



import clearskies.di


class MySQLConnectionDynamicProducer(clearskies.di.additional_config.AdditionalConfig):
    _producer_name = None
    _database_host = None
    _database_name = None

    def __init__(self, producer_name=None, database_host=None, database_name=None):
        self._producer_name = producer_name
        self._database_host = database_host
        self._database_name = database_name

    def provide_connection_details(self, environment, secrets):
        if not secrets:
            raise ValueError(
                "I was asked to connect to a database via an AKeyless dynamic producer, \
                 but AKeyless itself wasn't configured. \
                Try setting the AKeyless auth method via clearskies.secrets.akeyless_[jwt|saml|aws_iam]_auth()"
            )

        producer_name = (
            self._producer_name
            if self._producer_name is not None
            else environment.get("akeyless_mysql_dynamic_producer", silent=True)
        )
        if not producer_name:
            raise ValueError(
                "I was asked to connect to a database via an AKeyless dynamic producer, \
                but I wasn't told the path to the dynamic producer. \
                This can be set in an environment variable named 'akeyless_mysql_dynamic_producer'\
                or it can be set in the configuration via the producer_name kwarg."
            )
        database_name = (
            self._database_name if self._database_name is not None else environment.get("db_database", silent=True)
        )
        if not database_name:
            raise ValueError(
                "I was asked to connect to a database via an AKeyless dynamic producer, \
                but I wasn't told the name of the database. \
                This can be set in an environment variable named 'db_database' \
                or it can be set in the configuration via the database_name kwarg."
            )
        database_host = (
            self._database_host if self._database_host is not None else environment.get("db_host", silent=True)
        )
        if not database_host:
            raise ValueError(
                "I was asked to connect to a database via an AKeyless dynamic producer, \
                but I wasn't told the host name of the database. \
                This can be set in an environment variable named 'db_host' \
                or it can be set in the configuration via the database_host kwarg."
            )
        credentials = secrets.get_dynamic_secret(producer_name)

        return {
            "username": credentials["user"],
            "password": credentials["password"],
            "host": database_host,
            "database": database_name,
        }



import clearskies.di

from .jwks import Jwks


class AuthorizationPassThrough(Jwks):
    """
    Authentication class with pass through.

    This authentication class takes the authentication header from the incoming request and reflects
    it on outgoing requests.
    """

    """
    The input output helper
    """
    input_output = clearskies.di.inject.InputOutput()

    def headers(self, retry_auth=False):
        return {"Authorization": self.input_output.request_headers.get("authorization", True)}



from __future__ import annotations

from typing import TYPE_CHECKING, Self, overload

from clearskies.column import Column

if TYPE_CHECKING:
    from clearskies import Model


class String(Column):
    """
    A simple string column.

    ```python
    import clearskies


    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            Pet,
            writeable_column_names=["name"],
            readable_column_names=["id", "name"],
        ),
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl http://localhost:8080 -d '{"name": "Spot"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "e5b8417f-91bc-4fe5-9b64-04f571a7b10a",
            "name": "Spot"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl http://localhost:8080 -d '{"name": 10}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "value should be a string"
        }
    }

    ```
    """

    _allowed_search_operators = ["<=>", "!=", "<=", ">=", ">", "<", "=", "in", "is not null", "is null", "like"]
    _descriptor_config_map = None

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> str:
        pass

    def __get__(self, instance, cls):
        if instance is None:
            self.model_class = cls
            return self

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            instance.get_columns()

        if self.name not in instance._data:
            return None  # type: ignore

        if self.name not in instance._transformed_data:
            instance._transformed_data[self.name] = self.from_backend(instance._data[self.name])

        return instance._transformed_data[self.name]

    def __set__(self, instance: Model, value: str) -> None:
        instance._next_data[self.name] = value

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        return "value should be a string" if type(value) != str else ""



from .schema import Schema


class Array(Schema):
    item_definition = None
    _type = "array"

    def __init__(self, name, item_definition, value=None):
        super().__init__(name, example=None, value=value)
        self.item_definition = item_definition



import datetime
from typing import Any

import clearskies.configs
from clearskies.di import InjectableProperties, inject


class Akeyless(clearskies.Configurable, clearskies.di.InjectableProperties):
    requests = clearskies.di.inject.Requests()
    environment = clearskies.di.inject.Environment()
    akeyless = clearskies.di.inject.ByName("akeyless")

    access_id = clearskies.configs.String(required=True, regexp=r"^p-[\d\w]+$")
    access_type = clearskies.configs.Select(["aws_iam", "saml", "jwt"], required=True)
    api_host = clearskies.configs.String(default="https://api.akeyless.io")
    profile = clearskies.configs.String(regexp=r"^[\d\w\-]+$")

    _token_refresh: datetime.datetime = None  # type: ignore
    _token: str = ""
    _api: Any = None

    def __init__(self, access_id: str, access_type: str, jwt_env_key: str = "", api_host: str = "", profile: str = ""):
        self.access_id = access_id
        self.access_type = access_type
        self.jwt_env_key = jwt_env_key
        self.api_host = api_host
        self.profile = profile
        if self.access_type == "jwt" and not self.jwt_env_key:
            raise ValueError("When using the JWT access type for Akeyless you must provide jwt_env_key")

        self.finalize_and_validate_configuration()

    @property
    def api(self) -> Any:
        if self._api is None:
            configuration = self.akeyless.Configuration(host=self.api_host)
            self._api = self.akeyless.V2Api(self.akeyless.ApiClient(configuration))
        return self._api

    def create(self, path: str, value: Any) -> bool:
        res = self.api.create_secret(self.akeyless.CreateSecret(name=path, value=str(value), token=self._get_token()))
        return True

    def get(self, path: str, silent_if_not_found: bool = False) -> str:
        try:
            res = self._api.get_secret_value(self.akeyless.GetSecretValue(names=[path], token=self._get_token()))
        except Exception as e:
            if e.status == 404:  # type: ignore
                if silent_if_not_found:
                    return ""
                raise KeyError(f"Secret '{path}' not found")
            raise e
        return res[path]

    def get_dynamic_secret(self, path: str, args: dict[str, Any] | None = None) -> Any:
        kwargs = {
            "name": path,
            "token": self._get_token(),
        }
        if args:
            kwargs["args"] = args  # type: ignore

        return self._api.get_dynamic_secret_value(self.akeyless.GetDynamicSecretValue(**kwargs))

    def get_rotated_secret(self, path: str, args: dict[str, Any] | None = None) -> Any:
        kwargs = {
            "names": path,
            "token": self._get_token(),
        }
        if args:
            kwargs["args"] = args  # type: ignore

        res = self._api.get_rotated_secret_value(self.akeyless.GetRotatedSecretValue(**kwargs))
        return res

    def list_secrets(self, path: str) -> list[Any]:
        res = self._api.list_items(self.akeyless.ListItems(path=path, token=self._get_token()))
        if not res.items:
            return []

        return [item.item_name for item in res.items]

    def update(self, path: str, value: Any) -> None:
        res = self._api.update_secret_val(
            self.akeyless.UpdateSecretVal(name=path, value=str(value), token=self._get_token())
        )

    def upsert(self, path: str, value: Any) -> None:
        try:
            self.update(path, value)
        except Exception as e:
            self.create(path, value)

    def list_sub_folders(self, main_folder: str) -> list[str]:
        """Return the list of secrets/sub folders in the given folder."""
        items = self._api.list_items(self.akeyless.ListItems(path=main_folder, token=self._get_token()))

        # akeyless will return the absolute path and end in a slash but we only want the folder name
        main_folder_string_len = len(main_folder)
        return [sub_folder[main_folder_string_len:-1] for sub_folder in items.folders]

    def get_ssh_certificate(self, cert_issuer: str, cert_username: str, path_to_public_file: str) -> Any:
        with open(path_to_public_file, "r") as fp:
            public_key = fp.read()

        res = self._api.get_ssh_certificate(
            self.akeyless.GetSSHCertificate(
                cert_username=cert_username,
                cert_issuer_name=cert_issuer,
                public_key_data=public_key,
                token=self._get_token(),
            )
        )

        return res.data

    def _get_token(self) -> str:
        # AKeyless tokens live for an hour
        if self._token is not None and (self._token_refresh - datetime.datetime.now()).total_seconds() > 10:
            return self._token

        auth_method_name = f"auth_{self.access_type}"
        if not hasattr(self, auth_method_name):
            raise ValueError(f"Requested Akeyless authentication with unsupported auth method: '{self.access_type}'")

        self._token_refresh = datetime.datetime.now() + datetime.timedelta(hours=0.5)
        self._token = getattr(self, auth_method_name)()
        return self._token

    def auth_aws_iam(self):
        from akeyless_cloud_id import CloudId  # type: ignore

        res = self._api.auth(
            self.akeyless.Auth(access_id=self.access_id, access_type="aws_iam", cloud_id=CloudId().generate())
        )
        return res.token

    def auth_saml(self):
        import os
        from pathlib import Path

        os.system(f"akeyless list-items --profile {self.profile} --path /not/a/real/path > /dev/null 2>&1")
        home = str(Path.home())
        with open(f"{home}/.akeyless/.tmp_creds/{self.profile}-{self.access_id}", "r") as creds_file:
            credentials = creds_file.read()

        # and now we can turn that into a token
        response = self.requests.post(
            "https://rest.akeyless.io/",
            data={
                "cmd": "static-creds-auth",
                "access-id": self.access_id,
                "creds": credentials.strip(),
            },
        )
        return response.json()["token"]

    def auth_jwt(self):
        if not self.jwt_env_key:
            raise ValueError(
                "To use AKeyless JWT Auth, "
                "you must specify the name of the ENV key to load the JWT from when configuring AKeyless"
            )
        res = self._api.auth(
            self.akeyless.Auth(access_id=self.access_id, access_type="jwt", jwt=self.environment.get(self.jwt_env_key))
        )
        return res.token


class AkeylessSaml(Akeyless):
    def __init__(self, access_id: str, api_host: str = "", profile: str = ""):
        return super().__init__(access_id, "saml", api_host=api_host, profile=profile)


class AkeylessJwt(Akeyless):
    def __init__(self, access_id: str, jwt_env_key: str = "", api_host: str = "", profile: str = ""):
        return super().__init__(access_id, "jwt", jwt_env_key=jwt_env_key, api_host=api_host, profile=profile)


class AkeylessAwsIam(Akeyless):
    def __init__(self, access_id: str, api_host: str = ""):
        return super().__init__(access_id, "aws_iam", api_host=api_host)



import unittest


class TestBase(unittest.TestCase):
    def tearDown(self):
        import clearskies.backends.memory_backend

        clearskies.backends.memory_backend.MemoryBackend.clear_table_cache()



import re
from typing import Any, Callable

import clearskies.typing
from clearskies import configs, parameters_to_properties
from clearskies.columns.string import String


class Phone(String):
    """
    A string column that stores a phone number.

    The main difference between this and a plain string column is that this will validate that the string contains
    a phone number (containing only digits, dashes, spaces, plus sign, and parenthesis) of the appropriate length.
    When persisting the value to the backend, this column removes all non-digit characters.

    If you also set the usa_only flag to true then it will also validate that it is a valid US number containing
    9 digits and, optionally, a leading `1`.  Example:

    ```python
    import clearskies


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        phone = clearskies.columns.Phone(usa_only=True)


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            User,
            writeable_column_names=["name", "phone"],
            readable_column_names=["id", "name", "phone"],
        ),
    )
    wsgi()
    ```

    Which you can invoke:

    ```bash
    $ curl http://localhost:8080 -d '{"name":"John Doe", "phone": "+1 (555) 451-1234"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "e2b4bdad-b70f-4d44-a94c-0e265868b4d2",
            "name": "John Doe",
            "phone": "15554511234"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl http://localhost:8080 -d '{"name":"John Doe", "phone": "555 451-1234"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "aea34022-4b75-4eed-ac92-65fa4f4511ae",
            "name": "John Doe",
            "phone": "5554511234"
        },
        "pagination": {},
        "input_errors": {}
    }


    $ curl http://localhost:8080 -d '{"name":"John Doe", "phone": "555 451-12341"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "phone": "Invalid phone number"
        }
    }

    $ curl http://localhost:8080 -d '{"name":"John Doe", "phone": "1-2-3-4 asdf"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "phone": "Invalid phone number"
        }
    }
    ```
    """

    """ Whether or not to allow non-USA numbers. """
    usa_only = configs.Boolean(default=True)
    _descriptor_config_map = None

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        usa_only: bool = True,
        default: str | None = None,
        setable: str | Callable[..., str] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def to_backend(self, data: dict[str, Any]) -> dict[str, Any]:
        if not data.get(self.name):
            return data

        # phone numbers are stored as only digits.
        return {**data, **{self.name: re.sub(r"\D", "", data[self.name])}}

    def input_error_for_value(self, value: str, operator: str | None = None) -> str:
        if type(value) != str:
            return f"Value must be a string for {self.name}"

        # we'll allow spaces, dashes, parenthesis, dashes, and plus signs.
        # if there is anything else then it's not a valid phone number.
        # However, we don't do more detailed validation, because I'm too lazy to
        # figure out what is and is not a valid phone number, especially when
        # you get to the world of international numbers.
        if re.search(r"[^\d \-()+]", value):
            return "Invalid phone number"

        # for some final validation (especially US numbers) work only with the digits.
        value = re.sub(r"\D", "", value)

        if len(value) > 15:
            return "Invalid phone number"

        # we can't be too short unless we're doing a fuzzy search
        if len(value) < 10 and operator and operator.lower() != "like":
            return "Invalid phone number"

        if self.usa_only:
            if len(value) > 11:
                return "Invalid phone number"
            if value[0] == "1" and len(value) != 11:
                return "Invalid phone number"
            if value[0] != "1" and len(value) != 10:
                return "Invalid phone number"

        return ""



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING, Any, Callable

import requests

import clearskies.columns.datetime
import clearskies.columns.json
import clearskies.configs
import clearskies.configurable
import clearskies.model
import clearskies.query
from clearskies import parameters_to_properties
from clearskies.autodoc.schema import Integer as AutoDocInteger
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.autodoc.schema import String as AutoDocString
from clearskies.backends.backend import Backend
from clearskies.di import InjectableProperties, inject
from clearskies.functional import routing, string

if TYPE_CHECKING:
    import clearskies.column


class ApiBackend(clearskies.configurable.Configurable, Backend, InjectableProperties):
    """
    Fetch and store data from an API endpoint.

    The ApiBackend gives developers a way to quickly build SDKs to connect a clearskies applications
    to arbitrary API endpoints.  The backend has some built in flexibility to make it easy to connect it to
    **most** APIs, as well as behavioral hooks so that you can override small sections of the logic to accommodate
    APIs that don't work in the expected way.  This allows you to interact with APIs using the standard model
    methods, just like every other backend, and also means that you can attach such models to endpoints to
    quickly enable all kinds of pre-defined behaviors.

    ## Usage

    Configuring the API backend is pretty easy:

     1. Provide the `base_url` to the constructor, or extend it and set it in the `__init__` for the new backend.
     2. Provide a `clearskies.authentication.Authentication` object, assuming it isn't a public API.
     3. Match your model class name to the path of the API (or set `model.destination_name()` appropriately)
     4. Use the resulting model like you would any other model!

    It's important to understand how the Api Backend will map queries and saves to the API in question.  The rules
    are fairly simple:

      1. The API backend only supports searching with the equals operator (e.g. `models.where("column=value")`).
      2. To specify routing parameters, use the `{parameter_name}` or `:parameter_name` syntax in either the url
         or in the destination name of your model.  In order to query the model, you then **must** provide a value
         for any routing parameters, using a matching search condition: (e.g.
         `models.where("routing_parameter_name=value")`)
      3. Any search clauses that don't correspond to routing parameters will be translated into query parameters.
         So, if your destination_name is `https://example.com/:categoy_id/products` and you executed a
         model query: `models.where("category_id=10").where("on_sale=1")` then this would result in fetching
         a URL of `https://example.com/10/products?on_sale=1`
      4. When you specifically search on the id column for the model, the id will be appended to the end
         of the URL rather than as a query parameter.  So, with a destination name of `https://example.com/products`,
         querying for `models.find("id=10")` will result in fetching `https://example.com/products/10`.
      5. Delete and Update operations will similarly append the id to the URL, and also set the appropriate
         response method (e.g. `DELETE` or `PATCH` by default).
      6. When processing the response, the backend will attempt to automatically discover the results by looking
         for dictionaries that contain the expected column names (as determined from the model schema and the mapping
         rules).
      7. The backend will check for a response header called `link` and parse this to find pagination information
         so it can iterate through records.

    NOTE: The API backend doesn't support joins or group_by clauses.  This limitation, as well as the fact that it only
    supports seaching with the equals operator, isn't a limitation in the API backend itself, but simply reflects the behavior
    of most API endoints.  If you want to support an API that has more flexibility (for instance, perhaps it allows for more search
    operations than just `=`), then you can extend the appropritae methods, discussed below, to map a model query to an API request.

    Here's an example of how to use the API Backend to integrate with the Github API:

    ```python
    import clearskies


    class GithubPublicBackend(clearskies.backends.ApiBackend):
        def __init__(
            self,
            # This varies from endpoint to endpoint, so we want to be able to set it for each model
            pagination_parameter_name: str = "since",
        ):
            # these are fixed for all gitlab API parameters, so there's no need to make them setable
            # from the constructor
            self.base_url = "https://api.github.com"
            self.limit_parameter_name = "per_page"
            self.pagination_parameter_name = pagination_parameter_name
            self.finalize_and_validate_configuration()


    class UserRepo(clearskies.Model):
        # Corresponding API Docs: https://docs.github.com/en/rest/repos/repos?apiVersion=2022-11-28#list-repositories-for-a-user
        id_column_name = "full_name"
        backend = GithubPublicBackend(pagination_parameter_name="page")

        @classmethod
        def destination_name(cls) -> str:
            return "users/:login/repos"

        id = clearskies.columns.Integer()
        full_name = clearskies.columns.String()
        type = clearskies.columns.Select(["all", "owner", "member"])
        url = clearskies.columns.String()
        html_url = clearskies.columns.String()
        created_at = clearskies.columns.Datetime()
        updated_at = clearskies.columns.Datetime()

        # The API endpoint won't return "login" (e.g. username), so it may not seem like a column, but we need to search by it
        # because it's a URL parameter for this API endpoint.  Clearskies uses strict validation and won't let us search by
        # a column that doesn't exist in the model: therefore, we have to add the login column.
        login = clearskies.columns.String(is_searchable=True, is_readable=False)

        # The API endpoint let's us sort by `created`/`updated`.  Note that the names of the columns (based on the data returned
        # by the API endpoint) are `created_at`/`updated_at`.  As above, clearskies strictly validates data, so we need columns
        # named created/updated so that we can sort by them.  We can set some flags to (hopefully) avoid confusion
        updated = clearskies.columns.Datetime(
            is_searchable=False, is_readable=False, is_writeable=False
        )
        created = clearskies.columns.Datetime(
            is_searchable=False, is_readable=False, is_writeable=False
        )


    class User(clearskies.Model):
        # Corresponding API docs: https://docs.github.com/en/rest/users/users?apiVersion=2022-11-28#list-users

        # github has two columns that are both effecitvely id columns: id and login.
        # We use the login column for id_column_name because that is the column that gets
        # used in the API to fetch an individual record
        id_column_name = "login"
        backend = GithubPublicBackend()

        id = clearskies.columns.Integer()
        login = clearskies.columns.String()
        gravatar_id = clearskies.columns.String()
        avatar_url = clearskies.columns.String()
        html_url = clearskies.columns.String()
        repos_url = clearskies.columns.String()

        # We can hook up relationships between models just like we would if we were using an SQL-like
        # database.  The whole point of the backend system is that the model queries work regardless of
        # backend, so clearskies can issue API calls to fetch related records just like it would be able
        # to fetch children from a related database table.
        repos = clearskies.columns.HasMany(
            UserRepo,
            foreign_column_name="login",
            readable_child_columns=["id", "full_name", "html_url"],
        )


    def fetch_user(users: User, user_repos: UserRepo):
        # If we execute this models query:
        some_repos = (
            user_repos.where("login=cmancone")
            .sort_by("created", "desc")
            .where("type=owner")
            .pagination(page=2)
            .limit(5)
        )
        # the API backend will fetch this url:
        # https://api.github.com/users/cmancone/repos?type=owner&sort=created&direction=desc&per_page=5&page=2
        # and we can use the results like always
        repo_names = [repo.full_name for repo in some_repos]

        # For the below case, the backend will fetch this url:
        # https://api.github.com/users/cmancone
        # in addition, the readable column names on the callable endpoint includes "repos", which references our has_many
        # column.  This means that when converting the user model to JSON, it will also grab a page of repositories for that user.
        # To do that, it will fetch this URL:
        # https://api.github.com/users/cmancone/repos
        return users.find("login=cmancone")


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Callable(
            fetch_user,
            model_class=User,
            readable_column_names=["id", "login", "html_url", "repos"],
        ),
        classes=[User, UserRepo],
    )

    if __name__ == "__main__":
        wsgi()
    ```

    The following example demonstrates how models using this backend can be used in other clearskies endpoints, just like any
    other model.  Note that the following example is re-using the above models and backend, I have just omitted them for the sake
    of brevity:

    ```python
    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            model_class=User,
            readable_column_names=["id", "login", "html_url"],
            sortable_column_names=["id"],
            default_sort_column_name=None,
            default_limit=10,
        ),
        classes=[User],
    )

    if __name__ == "__main__":
        wsgi()
    ```

    And if you invoke it:

    ```bash
    $ curl 'http://localhost:8080' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": 1,
                "login": "mojombo",
                "html_url": "https://github.com/mojombo"
            },
            {
                "id": 2,
                "login": "defunkt",
                "html_url": "https://github.com/defunkt"
            },
            {
                "id": 3,
                "login": "pjhyett",
                "html_url": "https://github.com/pjhyett"
            },
            {
                "id": 4,
                "login": "wycats",
                "html_url": "https://github.com/wycats"
            },
            {
                "id": 5,
                "login": "ezmobius",
                "html_url": "https://github.com/ezmobius"
            },
            {
                "id": 6,
                "login": "ivey",
                "html_url": "https://github.com/ivey"
            },
            {
                "id": 7,
                "login": "evanphx",
                "html_url": "https://github.com/evanphx"
            },
            {
                "id": 17,
                "login": "vanpelt",
                "html_url": "https://github.com/vanpelt"
            },
            {
                "id": 18,
                "login": "wayneeseguin",
                "html_url": "https://github.com/wayneeseguin"
            },
            {
                "id": 19,
                "login": "brynary",
                "html_url": "https://github.com/brynary"
            }
        ],
        "pagination": {
            "number_results": null,
            "limit": 10,
            "next_page": {
                "since": "19"
            }
        },
        "input_errors": {}
    }
    ```

    In essence, we now have an endpoint that lists results but, instead of pulling its data from a database, it
    makes API calls.  It also tracks pagination as expected, so you can use the data in `pagination.next_page` to
    fetch the next set of results, just as you would if this were backed by a database, e.g.:

    ```bash
    $ curl http://localhost:8080?since=19
    ```

    ## Mapping from Queries to API calls

    The process of mapping a model query into an API request involves a few different methods which can be
    overwritten to fully control the process.  This is necessary in cases where an API behaves differently
    than expected by the API backend.  This table outlines the method involved and how they are used:

    | Method                           | Description                                                                                           |
    |----------------------------------|-------------------------------------------------------------------------------------------------------|
    | records_url                      | Return the absolute URL to fetch, as well as any columns that were used to fill in routing parameters |
    | records_method                   | Reurn the HTTP request method to use for the API call                                                 |
    | conditions_to_request_parameters | Translate the query conditions into URL fragments, query parameters, or JSON body parameters          |
    | pagination_to_request_parameters | Translate the pagination data into URL fragments, query parameters, or JSON body parameters           |
    | sorts_to_request_parameters      | Translate the sort directive(s) into URL fragments, query parameters, or JSON body parameters         |
    | map_records_response             | Take the response from the API and return a list of dictionaries with the resulting records           |

    In short, the details of the query are stored in a clearskies.query.Query object which is passed around to these
    various methods.  They use that information to adjust the URL, add query parameters, or add parameters into the
    JSON body.  The API Backend will then execute an API call with those final details, and use the map_record_response
    method to pull the returned records out of the response from the API endpoint.

    """

    can_count = False

    """
    The Base URL for the requests - will be prepended to the destination_name() from the model.

    Note: this is treated as a 'folder' path: if set, it becomes the URL prefix and is followed with a '/'
    """
    base_url = clearskies.configs.String(default="")

    """
    A suffix to append to the end of the URL.

    Note: this is treated as a 'folder' path: if set, it becomes the URL suffix and is prefixed with a '/'
    """
    url_suffix = clearskies.configs.String(default="")

    """
    An instance of clearskies.authentication.Authentication that handles authentication to the API.

    The following example is a modification of the Github Backends used above that shows how to setup authentication.
    Github, like many APIs, uses an API key attached to the request via the authorization header.  The SecretBearer
    authentication class in clearskies is designed for this common use case, and pulls the secret key out of either
    an environment variable or the secret manager (I use the former in this case, because it's hard to have a
    self-contained example with a secret manager).  Of course, any authentication method can be attached to your
    API backend - SecretBearer authentication is used here simply because it's a common approach.

    Note that, when used in conjunction with a secret manager, the API Backend and the SecretBearer class will work
    together to check for a new secret in the event of an authentication failure from the API endpoint (specifically,
    a 401 error).  This allows you to automate credential rotation: create a new API key, put it in the secret manager,
    and then revoke the old API key.  The next time an API call is made, the SecretBearer will provide the old key from
    it's cache and the request will fail.  The API backend will detect this and try the request again, but this time
    will tell the SecretBearer class to refresh it's cache with a fresh copy of the key from the secrets manager.
    Therefore, as long as you put the new key in your secret manager **before** disabling the old key, this second
    request will succeed and the service will continue to operate successfully with only a slight delay in response time
    caused by refreshing the cache.

    ```python
    import clearskies

    class GithubBackend(clearskies.backends.ApiBackend):
        def __init__(
            self,
            pagination_parameter_name: str = "page",
            authentication: clearskies.authentication.Authentication | None = None,
        ):
            self.base_url = "https://api.github.com"
            self.limit_parameter_name = "per_page"
            self.pagination_parameter_name = pagination_parameter_name
            self.authentication = clearskies.authentication.SecretBearer(
                environment_key="GITHUB_API_KEY",
                header_prefix="Bearer ", # Because github expects a header of 'Authorization: Bearer API_KEY'
            )
            self.finalize_and_validate_configuration()

    class Repo(clearskies.Model):
        id_column_name = "login"
        backend = GithubBackend()

        @classmethod
        def destination_name(cls):
            return "/user/repos"

        id = clearskies.columns.Integer()
        name = clearskies.columns.String()
        full_name = clearskies.columns.String()
        html_url = clearskies.columns.String()
        visibility = clearskies.columns.Select(["all", "public", "private"])

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            model_class=Repo,
            readable_column_names=["id", "name", "full_name", "html_url"],
            sortable_column_names=["full_name"],
            default_sort_column_name="full_name",
            default_limit=10,
            where=["visibility=private"],
        ),
        classes=[Repo],
    )

    if __name__ == "__main__":
        wsgi()

    ```
    """
    authentication = clearskies.configs.Authentication(default=None)

    """
    A dictionary of headers to attach to all outgoing API requests
    """
    headers = clearskies.configs.StringDict(default={})

    """
    The casing used in the model (snake_case, camelCase, TitleCase)

    This is used in conjunction with api_casing to tell the processing layer when you and the API are using
    different casing standards.  The API backend will then automatically covnert the casing style of the API
    to match your model.  This can be helpful when you have a standard naming convention in your own code which
    some external API doesn't follow, that way you can at least standardize things in your code.  In the following
    example, these parameters are used to convert from the snake_casing native to the Github API into the
    TitleCasing used in the model class:

    ```python
    import clearskies

    class User(clearskies.Model):
        id_column_name = "login"
        backend = clearskies.backends.ApiBackend(
            base_url="https://api.github.com",
            limit_parameter_name="per_page",
            pagination_parameter_name="since",
            model_casing="TitleCase",
            api_casing="snake_case",
        )

        Id = clearskies.columns.Integer()
        Login = clearskies.columns.String()
        GravatarId = clearskies.columns.String()
        AvatarUrl = clearskies.columns.String()
        HtmlUrl = clearskies.columns.String()
        ReposUrl = clearskies.columns.String()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            model_class=User,
            readable_column_names=["Login", "AvatarUrl", "HtmlUrl", "ReposUrl"],
            sortable_column_names=["Id"],
            default_sort_column_name=None,
            default_limit=2,
            internal_casing="TitleCase",
            external_casing="TitleCase",
        ),
        classes=[User],
    )

    if __name__ == "__main__":
        wsgi()
    ```

    and when executed:

    ```bash
    $ curl http://localhost:8080 | jq
    {
        "Status": "Success",
        "Error": "",
        "Data": [
            {
                "Login": "mojombo",
                "AvatarUrl": "https://avatars.githubusercontent.com/u/1?v=4",
                "HtmlUrl": "https://github.com/mojombo",
                "ReposUrl": "https://api.github.com/users/mojombo/repos"
            },
            {
                "Login": "defunkt",
                "AvatarUrl": "https://avatars.githubusercontent.com/u/2?v=4",
                "HtmlUrl": "https://github.com/defunkt",
                "ReposUrl": "https://api.github.com/users/defunkt/repos"
            }
        ],
        "Pagination": {
            "NumberResults": null,
            "Limit": 2,
            "NextPage": {
                "Since": "2"
            }
        },
        "InputErrors": {}
    }
    ```
    """
    model_casing = clearskies.configs.Select(["snake_case", "camelCase", "TitleCase"], default="snake_case")

    """
    The casing used by the API response (snake_case, camelCase, TitleCase)

    See model_casing for details and usage.
    """
    api_casing = clearskies.configs.Select(["snake_case", "camelCase", "TitleCase"], default="snake_case")

    """
    A mapping from the data keys returned by the API to the data keys expected in the model

    This comes into play when you want your model columns to use different names than what is returned by the
    API itself.  Provide a dictionary where the key is the name of a piece of data from the API, and the value
    is the name of the column in the model.  The API Backend will use this to match the API data to your model.
    In the example below, `html_url` from the API has been mapped to `profile_url` in the model:

    ```python
    import clearskies

    class User(clearskies.Model):
        id_column_name = "login"
        backend = clearskies.backends.ApiBackend(
            base_url="https://api.github.com",
            limit_parameter_name="per_page",
            pagination_parameter_name="since",
            api_to_model_map={"html_url": "profile_url"},
        )

        id = clearskies.columns.Integer()
        login = clearskies.columns.String()
        profile_url = clearskies.columns.String()

    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            model_class=User,
            readable_column_names=["login", "profile_url"],
            sortable_column_names=["id"],
            default_sort_column_name=None,
            default_limit=2,
        ),
        classes=[User],
    )

    if __name__ == "__main__":
        wsgi()
    ```

    And if you invoke it:

    ```bash
    $ curl http://localhost:8080 | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "login": "mojombo",
                "profile_url": "https://github.com/mojombo"
            },
            {
                "login": "defunkt",
                "profile_url": "https://github.com/defunkt"
            }
        ],
        "pagination": {
            "number_results": null,
            "limit": 2,
            "next_page": {
                "since": "2"
            }
        },
        "input_errors": {}
    }
    ```
    """
    api_to_model_map = clearskies.configs.StringDict(default={})

    """
    The name of the pagination parameter
    """
    pagination_parameter_name = clearskies.configs.String(default="start")

    """
    The expected 'type' of the pagination parameter: must be either 'int' or 'str'

    Note: this is set as a literal string, not as a type.
    """
    pagination_parameter_type = clearskies.configs.Select(["int", "str"], default="str")

    """
    The name of the parameter that sets the number of records per page (if empty, setting the page size will not be allowed)
    """
    limit_parameter_name = clearskies.configs.String(default="limit")

    """
    The requests instance.
    """
    requests = inject.Requests()

    """
    The dependency injection container (so we can pass it along to the Authentication object)
    """
    di = inject.Di()

    _auth_injected = False
    _response_to_model_map: dict[str, str] = None  # type: ignore

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        base_url: str,
        authentication: clearskies.authentication.Authentication | None = None,
        model_casing: str = "snake_case",
        api_casing: str = "snake_case",
        api_to_model_map: dict[str, str] = {},
        pagination_parameter_name: str = "start",
        pagination_parameter_type: str = "str",
        limit_parameter_name: str = "limit",
    ):
        self.finalize_and_validate_configuration()

    def finalize_url(self, url: str, available_routing_data: dict[str, str], operation: str) -> tuple[str, list[str]]:
        """
        Given a URL, this will append the base URL, fill in any routing data, and also return any used routing parameters.

        For example, consider a base URL of `/my/api/{record_id}/:other_id` and then this is called as so:

        ```python
        (url, used_routing_parameters) = api_backend.finalize_url(
            "entries",
            {
                "record_id": "1-2-3-4",
                "other_id": "a-s-d-f",
                "more_things": "qwerty",
            },
        )
        ```

        The returned url would be `/my/api/1-2-3-4/a-s-d-f/entries`, and used_routing_parameters would be ["record_id", "other_id"].
        The latter is returned so you can understand what parameters were absorbed into the URL.  Often, when some piece of data
        becomes a routing parameter, it needs to be ignored in the rest of the request.  `used_routing_parameters` helps with that.
        """
        base_url = self.base_url.strip("/") + "/" if self.base_url.strip("/") else ""
        url_suffix = "/" + self.url_suffix.strip("/") if self.url_suffix.strip("/") else ""
        url = base_url + url + url_suffix
        routing_parameters = routing.extract_url_parameter_name_map(url)
        if not routing_parameters:
            return (url, [])

        parts = url.split("/")
        used_routing_parameters = []
        for parameter_name, index in routing_parameters.items():
            if parameter_name not in available_routing_data:
                a = "an" if operation == "update" else "a"
                raise ValueError(
                    f"""Failed to generate URL while building {a} {operation} request!  Url {url} hsa a routing parameter named
                    {parameter_name} that I couldn't fill in from the request details.  When fetching records, this should be
                    provided by adding an equals condition to the model, e.g. `model.where("{parameter_name}=some_value")`.
                    When creating/updating a record, this should be provided in the save data, e.g.:
                    `model.save({{"{parameter_name}": "some_value"}})`
                    """
                )
            if available_routing_data[parameter_name].__class__ not in [str, int]:
                parameter_type = available_routing_data[parameter_name].__class__.__name__
                raise ValueError(
                    f"I was filling in a routing parameter named {parameter_name} but the value I was given has a type of {parameter_type}.  Routing parameters can only be strings or integers."
                )
            parts[index] = available_routing_data[parameter_name]
            used_routing_parameters.append(parameter_name)
        return ("/".join(parts), used_routing_parameters)

    def finalize_url_from_data(self, url: str, data: dict[str, Any], operation: str) -> tuple[str, list[str]]:
        """
        Create the final URL using a data dictionary to fill in any URL parameters.

        See finalize_url for more details about the return value
        """
        return self.finalize_url(url, data, operation)

    def finalize_url_from_query(self, query: clearskies.query.Query, operation: str) -> tuple[str, list[str]]:
        """
        Create the URL using a query to fill in any URL parameters.

        See finalize_url for more details about the return value
        """
        available_routing_data = {}
        for condition in query.conditions:
            if condition.operator != "=":
                continue
            available_routing_data[condition.column_name] = condition.values[0]
        return self.finalize_url(query.model_class.destination_name(), available_routing_data, operation)

    def create_url(self, data: dict[str, Any], model: clearskies.model.Model) -> tuple[str, list[str]]:
        """
        Calculate the URL to use for a create requst.  Also, return the list of ay data parameters used to construct the URL.

        See finalize_url for more details on the return value.
        """
        return self.finalize_url_from_data(model.destination_name(), data, "create")

    def create_method(self, data: dict[str, Any], model: clearskies.model.Model) -> str:
        """Return the request method to use with a create request."""
        return "POST"

    def records_url(self, query: clearskies.query.Query) -> tuple[str, list[str]]:
        """
        Calculate the URL to use for a records request.  Also, return the list of any query parameters used to construct the URL.

        See finalize_url for more details on the return value.
        """
        return self.finalize_url_from_query(query, "records")

    def records_method(self, query: clearskies.query.Query) -> str:
        """Return the request method to use when fetching records from the API."""
        return "GET"

    def count_url(self, query: clearskies.query.Query) -> tuple[str, list[str]]:
        """
        Calculate the URL to use for a request to get a record count..  Also, return the list of any query parameters used to construct the URL.

        See finalize_url for more details on the return value.
        """
        return self.records_url(query)

    def count_method(self, query: clearskies.query.Query) -> str:
        """Return the request method to use when making a request for a record count."""
        return self.records_method(query)

    def delete_url(self, id: int | str, model: clearskies.model.Model) -> tuple[str, list[str]]:
        """
        Calculate the URL to use for a delete request.  Also, return the list of any query parameters used to construct the URL.

        See finalize_url for more details on the return value.
        """
        model_base_url = model.destination_name().strip("/") + "/" if model.destination_name() else ""
        return self.finalize_url_from_data(f"{model_base_url}{id}", model.get_raw_data(), "delete")

    def delete_method(self, id: int | str, model: clearskies.model.Model) -> str:
        """Return the request method to use when deleting records via the API."""
        return "DELETE"

    def update_url(self, id: int | str, data: dict[str, Any], model: clearskies.model.Model) -> tuple[str, list[str]]:
        """
        Calculate the URL to use for an update request.  Also, return the list of any query parameters used to construct the URL.

        See finalize_url for more details on the return value.
        """
        model_base_url = model.destination_name().strip("/") + "/" if model.destination_name() else ""
        return self.finalize_url_from_data(f"{model_base_url}{id}", {**model.get_raw_data(), **data}, "update")

    def update_method(self, id: int | str, data: dict[str, Any], model: clearskies.model.Model) -> str:
        """Return the request method to use for an update request."""
        return "PATCH"

    def update(self, id: int | str, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        """Update a record."""
        data = {**data}
        (url, used_routing_parameters) = self.update_url(id, data, model)
        request_method = self.update_method(id, data, model)
        for parameter in used_routing_parameters:
            del data[parameter]

        response = self.execute_request(url, request_method, json=data)
        json_response = response.json() if response.content else {}
        new_record = {**model.get_raw_data(), **data}
        if response.content:
            new_record = {**new_record, **self.map_update_response(response.json(), model)}
        return new_record

    def map_update_response(self, response_data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        """
        Take the response from the API endpoint for an update request and figure out where the data lives/return it to build a new model.

        See self.map_record_response for goals/motiviation
        """
        return self.map_record_response(response_data, model.get_columns(), "update")

    def create(self, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        """Create a record."""
        data = {**data}
        (url, used_routing_parameters) = self.create_url(data, model)
        request_method = self.create_method(data, model)
        for parameter in used_routing_parameters:
            del data[parameter]

        response = self.execute_request(url, request_method, json=data, headers=self.headers)
        json_response = response.json() if response.content else {}
        if response.content:
            return self.map_create_response(response.json(), model)
        return {}

    def map_create_response(self, response_data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        return self.map_record_response(response_data, model.get_columns(), "create")

    def delete(self, id: int | str, model: clearskies.model.Model) -> bool:
        (url, used_routing_parameters) = self.delete_url(id, model)
        request_method = self.delete_method(id, model)

        response = self.execute_request(url, request_method)
        return True

    def records(
        self, query: clearskies.query.Query, next_page_data: dict[str, str | int] | None = None
    ) -> list[dict[str, Any]]:
        self.check_query(query)
        (url, method, body, headers) = self.build_records_request(query)
        response = self.execute_request(url, method, json=body, headers=headers)
        records = self.map_records_response(response.json(), query)
        if isinstance(next_page_data, dict):
            self.set_next_page_data_from_response(next_page_data, query, response)
        return records

    def build_records_request(self, query: clearskies.query.Query) -> tuple[str, str, dict[str, Any], dict[str, str]]:
        (url, used_routing_parameters) = self.records_url(query)

        (condition_route_id, condition_url_parameters, condition_body_parameters) = (
            self.conditions_to_request_parameters(query, used_routing_parameters)
        )
        (pagination_url_parameters, pagination_body_parameters) = self.pagination_to_request_parameters(query)
        (sort_url_parameters, sort_body_parameters) = self.sorts_to_request_parameters(query)

        url_parameters = {
            **condition_url_parameters,
            **pagination_url_parameters,
            **sort_url_parameters,
        }

        body_parameters = {
            **condition_body_parameters,
            **pagination_body_parameters,
            **sort_body_parameters,
        }

        if condition_route_id:
            url = url.rstrip("/") + "/" + condition_route_id
        if url_parameters:
            url = url + "?" + urllib.parse.urlencode(url_parameters)

        return (
            url,
            self.records_method(query),
            body_parameters,
            {},
        )

    def conditions_to_request_parameters(
        self, query: clearskies.query.Query, used_routing_parameters: list[str]
    ) -> tuple[str, dict[str, str], dict[str, Any]]:
        route_id = ""

        url_parameters = {}
        for condition in query.conditions:
            if condition.column_name in used_routing_parameters:
                continue
            if condition.operator != "=":
                raise ValueError(
                    f"I'm not very smart and only know how to search with the equals operator, but I received a condition of {condition.parsed}.  If you need to support this, you'll have to extend the ApiBackend and overwrite the build_records_request method."
                )
            if condition.column_name == query.model_class.id_column_name:
                route_id = condition.values[0]
                continue
            url_parameters[condition.column_name] = condition.values[0]

        return (route_id, url_parameters, {})

    def pagination_to_request_parameters(self, query: clearskies.query.Query) -> tuple[dict[str, str], dict[str, Any]]:
        url_parameters = {}
        if query.limit:
            if not self.limit_parameter_name:
                raise ValueError(
                    "The records query attempted to change the limit (the number of results per page) but the backend does not support it.  If it actually does support this, then set an appropriate value for backend.limit_parameter_name"
                )
            url_parameters[self.limit_parameter_name] = str(query.limit)

        if query.pagination.get(self.pagination_parameter_name):
            url_parameters[self.pagination_parameter_name] = str(query.pagination.get(self.pagination_parameter_name))

        return (url_parameters, {})

    def sorts_to_request_parameters(self, query: clearskies.query.Query) -> tuple[dict[str, str], dict[str, Any]]:
        if not query.sorts:
            return ({}, {})

        if len(query.sorts) > 1:
            raise ValueError(
                "I received a query with two sort directives, but I can only handle one.  Sorry!  If you need o support two sort directions, you'll have to extend the ApiBackend and overwrite the build_records_request method."
            )

        return (
            {"sort": query.sorts[0].column_name, "direction": query.sorts[0].direction.lower()},
            {},
        )

    def map_records_response(
        self, response_data: Any, query: clearskies.query.Query, query_data: dict[str, Any] | None = None
    ) -> list[dict[str, Any]]:
        """Take the response from an API endpoint that returns a list of records and find the actual list of records."""
        columns = query.model_class.get_columns()
        # turn all of our conditions into record data and inject these into the results.  We do this to keep around
        # any query parameters.  This is especially important for any URL parameters, wihch aren't always returned in
        # the data, but which we are likely to need again if we go to update/delete the record.
        if query_data is None:
            query_data = {}
            for condition in query.conditions:
                if condition.operator != "=":
                    continue
                query_data[condition.column_name] = condition.values[0]

        # if our response is actually a list, then presumably the problem is solved.  If the response is a list
        # and the individual items aren't model results though... well, then I'm very confused
        if isinstance(response_data, list):
            if not response_data:
                return []
            if not self.check_dict_and_map_to_model(response_data[0], columns, query_data):
                raise ValueError(
                    f"The response from a records request returned a list, but the records in the list didn't look anything like the model class.  Please check your model class and mapping settings in the API Backend.  If those are correct, then you'll have to override the map_records_response method, because the API you are interacting with is returning data in an unexpected way that I can't automatically figure out."
                )
            return [self.check_dict_and_map_to_model(record, columns, query_data) for record in response_data]  # type: ignore

        if not isinstance(response_data, dict):
            raise ValueError(
                f"The response from a records request returned a variable of type {response_data.__class__.__name__}, which is just confusing.  To do automatic introspection, I need a list or a dictionary.  I'm afraid you'll have to extend the API backend and override the map_record_response method to deal with this."
            )

        for key, value in response_data.items():
            if not isinstance(value, list):
                continue
            return self.map_records_response(value, query, query_data)

        # a records request may only return a single record, so before we fail, let's check for that
        record = self.check_dict_and_map_to_model(response_data, columns, query_data)
        if record is not None:
            return [record]

        raise ValueError(
            "The response from a records request returned a dictionary, but none of the items in the dictionary was a list, so I don't know where to find the records.  I only ever check one level deep in dictionaries.  I'm afraid you'll have to extend the API backend and override the map_records_response method to deal with this."
        )

    def map_record_response(
        self, response_data: dict[str, Any], columns: dict[str, clearskies.column.Column], operation: str
    ) -> dict[str, Any]:
        """
        Take the response from an API endpoint that returns a single record (typically update and create requests) and return the data for a new model.

        The goal of this method is to try to use the model schema to automatically understand the response from the
        the API endpoint.  The goal is for the backend to work out-of-the-box with most APIs.  In general, it works
        by iterating over the response, looking for a dictionary with keys that match the expected model columns.

        Occassionally the automatic introspection may not be able to make sense of the response from an API
        endoint.  If this happens, you have to make a new API backend, override the map_record_response method
        to manage the mapping yourself, and then attach this new backend to your models.
        """
        an = "a" if operation == "create" else "an"
        if not isinstance(response_data, dict):
            raise ValueError(
                f"The response from {an} {operation} request returned a variable of type {response_data.__class__.__name__}, which is just confusing.  To do automatic introspection, I need a dictionary.  I'm afraid you'll have to build your own API backend and override the map_record_response method to deal with this."
            )

        response = self.check_dict_and_map_to_model(response_data, columns)
        if response is None:
            raise ValueError(
                f"I was not able to automatically interpret the response from {an} {operation} request.  This could be a sign of a response that is structured in a very unusual way, or may be a sign that the casing settings and/or columns on your model to properly reflect the API response.  For the former, you will hvae to build your own API backend and override the map_record_response to deal with this."
            )

        return response

    def check_dict_and_map_to_model(
        self,
        response_data: dict[str, Any],
        columns: dict[str, clearskies.column.Column],
        query_data: dict[str, Any] = {},
    ) -> dict[str, Any] | None:
        """
        Check a dictionary in the response to decide if it contains the data for a record.

        If not, it will search the keys for something that looks like a record.
        """
        # first let's get a coherent map of expected-key-names in the response to model names
        response_to_model_map = self.build_response_to_model_map(columns)

        # and now we can see if that appears to be what we have
        response_keys = set(response_data.keys())
        map_keys = set(response_to_model_map.keys())
        matching = response_keys.intersection(map_keys)

        # if nothing matches then clearly this isn't what we're looking for: repeat on all the children
        if not matching:
            for key, value in response_data.items():
                if not isinstance(value, dict):
                    continue
                mapped = self.check_dict_and_map_to_model(value, columns)
                if mapped:
                    return {**query_data, **mapped}

            # no match anywhere :(
            return None

        # we may need to be smarter about whether or not we think we found a match, but for now let's
        # ignore that possibility.  If any columns match between the keys in our response dictionary and
        # the keys that we are expecting to find data in, then just assume that we have found a record.
        mapped = {response_to_model_map[key]: response_data[key] for key in matching}

        # finally, move over anything not mentioned in the map
        for key in response_keys.difference(map_keys):
            mapped[string.swap_casing(key, self.api_casing, self.model_casing)] = response_data[key]

        return {**query_data, **mapped}

    def build_response_to_model_map(self, columns: dict[str, clearskies.column.Column]) -> dict[str, str]:
        if self._response_to_model_map is not None:
            return self._response_to_model_map

        self._response_to_model_map = {}
        for column_name in columns:
            self._response_to_model_map[string.swap_casing(column_name, self.model_casing, self.api_casing)] = (
                column_name
            )
        self._response_to_model_map = {**self._response_to_model_map, **self.api_to_model_map}

        return self._response_to_model_map

    def set_next_page_data_from_response(
        self,
        next_page_data: dict[str, Any],
        query: clearskies.query.Query,
        response: requests.Response,  # type: ignore
    ) -> None:
        """
        Update the next_page_data dictionary with the appropriate data needed to fetch the next page of records.

        This method has a very important job, which is to inform clearskies about how to make another API call to fetch the next
        page of records.  The way this happens is by updating the `next_page_data` dictionary in place with whatever pagination
        information is necessary.  Note that this relies on next_page_data being passed by reference, hence the need to update
        it in place.  That means that you can do this:

        ```python
        next_page_data["some_key"] = "some_value"
        ```

        but if you do this:

        ```python
        next_page_data = {"some_key": "some_value"}
        ```

        Then things simply won't work.
        """
        # Different APIs generally have completely different ways of communicating pagination data, but one somewhat common
        # approach is to use a link header, so let's support that in the base class.
        if "link" not in response.headers:
            return
        next_link = [rel for rel in response.headers["link"].split(",") if 'rel="next"' in rel]
        if not next_link:
            return
        parsed_next_link = urllib.parse.urlparse(next_link[0].split(";")[0].strip(" <>"))
        query_parameters = urllib.parse.parse_qs(parsed_next_link.query)
        if self.pagination_parameter_name not in query_parameters:
            raise ValueError(
                f"Configuration error with {self.__class__.__name__}!  I am configured to expect a pagination key of '{self.pagination_parameter_name}.  However, when I was parsing the next link from a response to get the next pagination details, I could not find the designated pagination key.  This likely means that backend.pagination_parameter_name is set to the wrong value.  The link in question was "
                + parsed_next_link.geturl()
            )
        next_page_data[self.pagination_parameter_name] = query_parameters[self.pagination_parameter_name][0]

    def count(self, query: clearskies.query.Query) -> int:
        raise NotImplementedError(
            f"The {self.__class__.__name__} backend does not support count operations, so you can't use the `len` or `bool` function for any models using it."
        )

    def execute_request(
        self,
        url: str,
        method: str,
        json: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        is_retry=False,
    ) -> requests.models.Response:  # type: ignore
        """
        Execute the actual API request and returns the response object.

        We don't directly call the requests library to support retries in the event of failed authentication.  The goal
        is to support short-lived credentials, and our authentication classes denote if they support this feature.  If
        they do, and the requests fails, then we'll ask the authentication method to refresh its credentials and we
        will retry the request.
        """
        if json is None:
            json = {}
        if headers is None:
            headers = {}

        if self.authentication:
            if not self._auth_injected:
                self._auth_injected = True
                if hasattr(self.authentication, "injectable_properties"):
                    self.authentication.injectable_properties(self.di)
            if is_retry:
                self.authentication.clear_credential_cache()
        # the requests library seems to build a slightly different request if you specify the json parameter,
        # even if it is null, and this causes trouble for some picky servers
        if not json:
            response = self.requests.request(
                method,
                url,
                headers=headers,
                auth=self.authentication if self.authentication else None,
            )
        else:
            response = self.requests.request(
                method,
                url,
                headers=headers,
                json=json,
                auth=self.authentication if self.authentication else None,
            )

        if not response.ok:
            if not is_retry and response.status_code == 401:
                return self.execute_request(url, method, json=json, headers=headers, is_retry=True)
            if not response.ok:
                raise ValueError(
                    f"Failed request.  Status code: {response.status_code}, message: "
                    + response.content.decode("utf-8")
                )

        return response

    def check_query(self, query: clearskies.query.Query) -> None:
        for key in ["joins", "group_by", "selects"]:
            if getattr(query, key):
                raise ValueError(f"{self.__class__.__name__} does not support queries with {key}")

        for condition in query.conditions:
            if condition.operator != "=":
                raise ValueError(
                    f"{self.__class__.__name__} only supports searching with the '=' operator, but I found a search with the {condition.operator} operator"
                )

    def validate_pagination_data(self, data: dict[str, Any], case_mapping: Callable) -> str:
        extra_keys = set(data.keys()) - set(self.allowed_pagination_keys())
        if len(extra_keys):
            key_name = case_mapping(self.pagination_parameter_name)
            return "Invalid pagination key(s): '" + "','".join(extra_keys) + f"'.  Only '{key_name}' is allowed"
        if self.pagination_parameter_name not in data:
            key_name = case_mapping(self.pagination_parameter_name)
            return f"You must specify '{key_name}' when setting pagination"
        value = data[self.pagination_parameter_name]
        try:
            if self.pagination_parameter_type == "int":
                converted = int(value)
        except:
            key_name = case_mapping(self.pagination_parameter_name)
            return f"Invalid pagination data: '{key_name}' must be a number"
        return ""

    def allowed_pagination_keys(self) -> list[str]:
        return [self.pagination_parameter_name]

    def documentation_pagination_next_page_response(self, case_mapping: Callable) -> list[Any]:
        if self.pagination_parameter_type == "int":
            return [AutoDocInteger(case_mapping(self.pagination_parameter_name), example=0)]
        else:
            return [AutoDocString(case_mapping(self.pagination_parameter_name), example="")]

    def documentation_pagination_next_page_example(self, case_mapping: Callable) -> dict[str, Any]:
        return {case_mapping(self.pagination_parameter_name): 0 if self.pagination_parameter_type == "int" else ""}

    def documentation_pagination_parameters(self, case_mapping: Callable) -> list[tuple[AutoDocSchema, str]]:
        return [
            (
                AutoDocInteger(
                    case_mapping(self.pagination_parameter_name),
                    example=0 if self.pagination_parameter_type == "int" else "",
                ),
                "The next record",
            )
        ]

    def column_from_backend(self, column: clearskies.column.Column, value: Any) -> Any:
        """We have a couple columns we want to override transformations for."""
        # most importantly, there's no need to transform a JSON column in either direction
        if isinstance(column, clearskies.columns.json.Json):
            return value
        return super().column_from_backend(column, value)

    def column_to_backend(self, column: clearskies.column.Column, backend_data: dict[str, Any]) -> dict[str, Any]:
        """We have a couple columns we want to override transformations for."""
        # most importantly, there's no need to transform a JSON column in either direction
        if isinstance(column, clearskies.columns.json.Json):
            return backend_data
        # also, APIs tend to have a different format for dates than SQL
        if isinstance(column, clearskies.columns.datetime.Datetime) and column.name in backend_data:
            as_date = (
                backend_data[column.name].isoformat()
                if type(backend_data[column.name]) != str
                else backend_data[column.name]
            )
            return {**backend_data, **{column.name: as_date}}
        return column.to_backend(backend_data)



from __future__ import annotations

from typing import TYPE_CHECKING

from clearskies import action
from clearskies.configs import config

if TYPE_CHECKING:
    from clearskies import typing


class Actions(config.Config):
    """
    Action config.

    A config that accepts various things that are accepted as actions in model lifecycle hooks:

     1. A callable (which should accept `model` as a parameter)
     2. An instance of clearskies.actions.Action
     3. A list containing any combination of the above

    Incoming values are normalized to a list so that a list always comes out even if a non-list is provided.
    """

    def __set__(self, instance, value: typing.action | list[typing.action]):
        if not isinstance(value, list):
            value = [value]

        for index, item in enumerate(value):
            if callable(item) or isinstance(item, action.Action):
                continue

            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{item.__class__.__name__}' for item #{index + 1} when a callable or Action is required"
            )

        instance._set_config(self, [*value])

    def __get__(self, instance, parent) -> list[typing.action]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



import datetime
import types
import uuid

from clearskies.di.injectable import Injectable


class Uuid(Injectable):
    def __init__(self, cache: bool = True):
        self.cache = cache

    def __get__(self, instance, parent) -> types.ModuleType:
        if instance is None:
            return self  # type: ignore
        return self._di.build_from_name("uuid", cache=self.cache)  # type: ignore



from clearskies.input_outputs.cli import Cli
from clearskies.input_outputs.headers import Headers
from clearskies.input_outputs.input_output import InputOutput
from clearskies.input_outputs.programmatic import Programmatic
from clearskies.input_outputs.wsgi import Wsgi

__all__ = [
    "Cli",
    "Headers",
    "InputOutput",
    "Programmatic",
    "Wsgi",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Self, overload

from clearskies.columns.category_tree_children import CategoryTreeChildren

if TYPE_CHECKING:
    from clearskies import Model


class CategoryTreeAncestors(CategoryTreeChildren):
    """
    A column to fetch the ancestors from a category tree column.

    See the CategoryTree column for usage examples.

    The ancestors are all parents of a given category, starting from the root category and working
    down to the direct parent.  So, given the following category tree:

    ```
    Root/
    ├─ Sub/
    │  ├─ Sub Sub/
    │  │  ├─ Sub Sub Sub/
    ├─ Another Child/
    ```

    The ancesotrs of `Sub Sub Sub` are `["Root", "Sub", "Sub Sub"]` while the ancestors of `Another Child`
    are `["Root"]`
    """

    _descriptor_config_map = None

    @overload
    def __get__(self, instance: None, cls: type) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type) -> Model:
        pass

    def __get__(self, model, cls):
        if model is None:
            self.model_class = cls
            return self  # type:  ignore

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            model.get_columns()

        return self.relatives(model, find_parents=True, include_all=True)



from typing import Any


class Object:
    def __init__(self, schema, oai3_schema_resolver):
        self.schema = schema
        self.oai3_schema_resolver = oai3_schema_resolver

    def convert(self, include_required=False):
        schema: dict[str, Any] = {
            "type": "object",
        }

        if self.schema.model_name:
            schema["$ref"] = f"#/components/schemas/{self.schema.model_name}"
        else:
            schema["properties"] = {
                schematic.name: self.oai3_schema_resolver(schematic).convert() for schematic in self.schema.children
            }

        if include_required:
            required = self.required()
            if required:
                schema["required"] = required

        return schema

    def required(self):
        return [schematic.name for schematic in self.schema.children if schematic.required]



import inspect
from functools import cmp_to_key
from typing import Any, Callable

import clearskies.model
import clearskies.query
from clearskies import functional
from clearskies.autodoc.schema import Integer as AutoDocInteger
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.backends.backend import Backend
from clearskies.di import InjectableProperties, inject


class Null:
    def __lt__(self, other):
        return True

    def __gt__(self, other):
        return False

    def __eq__(self, other):
        return isinstance(other, Null) or other is None


# for some comparisons we prefer comparing floats, but we need to be able to
# fall back on string comparison
def gentle_float_conversion(value):
    try:
        return float(value)
    except:
        return value


def _sort(row_a: Any, row_b: Any, sorts: list[clearskies.query.Sort], default_table_name: str) -> int:
    for sort in sorts:
        # so, if we've done a join then the rows will have data from all joined tables via a dict of dicts.
        # if there wasn't a join then we'll just have the data
        if sort.table_name in row_a and isinstance(row_a[sort.table_name], dict):
            sort_data_a = row_a[sort.table_name]
        elif not sort.table_name and default_table_name in row_a and isinstance(row_a[default_table_name], dict):
            sort_data_a = row_a[default_table_name]
        else:
            sort_data_a = row_a

        if sort.table_name in row_b and isinstance(row_b[sort.table_name], dict):
            sort_data_b = row_b[sort.table_name]
        elif not sort.table_name and default_table_name in row_b and isinstance(row_b[default_table_name], dict):
            sort_data_b = row_b[default_table_name]
        else:
            sort_data_b = row_b

        reverse = 1 if sort.direction.lower() == "asc" else -1
        value_a = sort_data_a[sort.column_name] if sort.column_name in sort_data_a else None
        value_b = sort_data_b[sort.column_name] if sort.column_name in sort_data_b else None
        if value_a == value_b:
            continue
        if value_a is None:
            return -1 * reverse
        if value_b is None:
            return 1 * reverse
        return reverse * (1 if value_a > value_b else -1)
    return 0


def cheating_equals(column, values, null):
    """
    Cheating because this solves a very specific problem that likely is a generic issue.

    The memory backend has some matching failures because boolean columns stay boolean in the
    memory store, but the incoming search values are not converted to boolean and tend to be
    str(1) or str(0).  The issue is that save data goes through the `to_backend` flow, but search
    data doesn't.  This doesn't matter most of the time because, in practice, the backend itself
    often does its own type conversion, but it causes problems for the memory backend.  I can't
    decide if fixing this will cause more problems than it solves, so for now I'm just cheating
    and putting in a hack for this specific use case :shame:.
    """

    def inner(row):
        backend_value = row[column] if column in row else null
        if isinstance(backend_value, bool):
            return backend_value == bool(values[0])
        return str(backend_value) == str(values[0])

    return inner


class MemoryTable:
    _table_name: str = ""
    _column_names: list[str] = []
    _rows: list[dict[str, Any] | None] = []
    null: Null = Null()
    _id_index: dict[int | str, int] = {}
    id_column_name: str = ""
    _next_id: int = 1
    _model_class: type[clearskies.model.Model] = None  # type:  ignore

    # here be dragons.  This is not a 100% drop-in replacement for the equivalent SQL operators
    # https://codereview.stackexchange.com/questions/259198/in-memory-table-filtering-in-python
    _operator_lambda_builders = {
        "<=>": lambda column, values, null: lambda row: row.get(column, null) == values[0],
        "!=": lambda column, values, null: lambda row: row.get(column, null) != values[0],
        "<=": (
            lambda column, values, null: lambda row: gentle_float_conversion(row.get(column, null))
            <= gentle_float_conversion(values[0])
        ),
        ">=": (
            lambda column, values, null: lambda row: gentle_float_conversion(row.get(column, null))
            >= gentle_float_conversion(values[0])
        ),
        ">": (
            lambda column, values, null: lambda row: gentle_float_conversion(row.get(column, null))
            > gentle_float_conversion(values[0])
        ),
        "<": (
            lambda column, values, null: lambda row: gentle_float_conversion(row.get(column, null))
            < gentle_float_conversion(values[0])
        ),
        "=": cheating_equals,
        "is not null": lambda column, values, null: lambda row: (column in row and row[column] is not None),
        "is null": lambda column, values, null: lambda row: (column not in row or row[column] is None),
        "is not": lambda column, values, null: lambda row: row.get(column, null) != values[0],
        "is": lambda column, values, null: lambda row: row.get(column, null) == str(values[0]),
        "like": lambda column, values, null: lambda row: row.get(column, null) == str(values[0]),
        "in": lambda column, values, null: lambda row: row.get(column, null) in values,
    }

    def __init__(self, model_class: type[clearskies.model.Model]) -> None:
        self._rows = []
        self._id_index = {}
        self.id_column_name = model_class.id_column_name
        self._next_id = 1
        self._model_class = model_class

        self._table_name = model_class.destination_name()
        self._column_names = list(model_class.get_columns().keys())
        if self.id_column_name not in self._column_names:
            self._column_names.append(self.id_column_name)

    def update(self, id: int | str, data: dict[str, Any]) -> dict[str, Any]:
        if id not in self._id_index:
            raise ValueError(f"Attempt to update non-existent record with '{self.id_column_name}' of '{id}'")
        index = self._id_index[id]
        row = self._rows[index]
        if row is None:
            raise ValueError(
                f"Cannot update record with '{self.id_column_name}' of '{id}' because it was already deleted"
            )
        for column_name in data.keys():
            if column_name not in self._column_names:
                raise ValueError(
                    f"Cannot update record: column '{column_name}' does not exist in table '{self._table_name}'"
                )
        self._rows[index] = {
            **self._rows[index],  # type: ignore
            **data,
        }
        return self._rows[index]  # type: ignore

    def create(self, data: dict[str, Any]) -> dict[str, Any]:
        for column_name in data.keys():
            if column_name not in self._column_names:
                raise ValueError(
                    f"Cannot create record: column '{column_name}' does not exist for model '{self._model_class.__name__}'"
                )
        incoming_id = data.get(self.id_column_name)
        if not incoming_id:
            incoming_id = self._next_id
            data[self.id_column_name] = incoming_id
            self._next_id += 1
        try:
            incoming_as_int = int(incoming_id)
            if incoming_as_int >= self._next_id:
                self._next_id = incoming_as_int + 1
        except:
            pass
        if incoming_id in self._id_index and self._rows[self._id_index[data[self.id_column_name]]] is not None:
            return self.update(data[self.id_column_name], data)
        for column_name in self._column_names:
            if column_name not in data:
                data[column_name] = None
        self._rows.append({**data})
        self._id_index[data[self.id_column_name]] = len(self._rows) - 1
        return data

    def delete(self, id):
        if id not in self._id_index:
            return True
        index = self._id_index[id]
        if self._rows[index] is None:
            return True
        # we set the row to None because if we remove it we'll change the indexes of the rest
        # of the rows, and I like being able to calculate the index from the id
        self._rows[index] = None
        return True

    def count(self, query: clearskies.query.Query):
        return len(self.rows(query, query.conditions, filter_only=True))

    def rows(
        self,
        query: clearskies.query.Query,
        conditions: list[clearskies.query.Condition],
        filter_only: bool = False,
        next_page_data: dict[str, Any] | None = None,
    ):
        rows = [row for row in self._rows if row is not None]
        for condition in conditions:
            rows = list(filter(self._condition_as_filter(condition), rows))
        rows = [*rows]
        if filter_only:
            return rows
        if query.sorts:
            rows = sorted(
                rows,
                key=cmp_to_key(
                    lambda row_a, row_b: _sort(row_a, row_b, query.sorts, query.model_class.destination_name())
                ),
            )
        if query.limit or query.pagination.get("start"):
            number_rows = len(rows)
            start = int(query.pagination.get("start", 0))
            if not start:
                start = 0
            if int(start) >= number_rows:
                start = number_rows - 1
            end = len(rows)
            if query.limit and start + int(query.limit) <= number_rows:
                end = start + int(query.limit)
            if end < number_rows and type(next_page_data) == dict:
                next_page_data["start"] = start + int(query.limit)
            rows = rows[start:end]
        return rows

    @classmethod
    def _condition_as_filter(cls, condition: clearskies.query.Condition) -> Callable:
        column = condition.column_name
        values = condition.values
        return cls._operator_lambda_builders[condition.operator.lower()](column, values, cls.null)


class MemoryBackend(Backend, InjectableProperties):
    """
    Store data in an in-memory store built in to clearskies.

    ## Usage

    Since the memory backend is built into clearskies, there's no configuration necessary to make it work:
    simply attach it to any of your models and it will manage data for you.  If you want though, you can declare
    a binding named `memory_backend_default_data` which you fill with records for your models to pre-populate
    the memory backend.  This can be helpful for tests as well as tables with fixed values.

    ### Testing

    A primary use case of the memory backend is for building unit tests of your code.  You can use the dependency
    injection system to override other backends with the memory backend.  You can still operate with model classes
    in the exact same way, so this can be an easy way to mock out databases/api endpoints/etc...  Of course,
    there can be behavioral differences between the memory backend and other backends, so this isn't always perfect.
    Hence why this works well for unit tests, but can't replace all testing, especially integration tests or
    end-to-end tests.  Here's an example of that.  Note that the UserPreference model uses the cursor backend,
    but when you run this code it will actually end up with the memory backend, so the code will run even without
    attempting to connect to a database.

    ```python
    import clearskies


    class UserPreference(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.CursorBackend()
        id = clearskies.columns.Uuid()


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            lambda user_preferences: user_preferences.create(no_data=True).id,
        ),
        classes=[UserPreference],
        class_overrides={
            clearskies.backends.CursorBackend: clearskies.backends.MemoryBackend(),
        },
    )

    if __name__ == "__main__":
        cli()
    ```

    Note that the model requests the CursorBackend, but then we switch that for the memory backend via the
    `class_overrides` kwarg to `clearskies.contexts.Cli`.  Therefore, the above code works regardless of
    whether or not a database is running.  Since the models are used to interact with the backend
    (rather than using the cursor directly), the above code works the same despite the change in backend.

    Again, this is helpful as a quick way to manage tests - swap out a database backend (or any other backend)
    for the memory backend, and then pre-populate records to test your application logic.  Obviously, this
    won't cach backend-specific issues (e.g. forgetting to add a column to your database, mismatches between
    column schema and backend schema, missing indexes, etc...), which is why this helps for unit tests
    but not for integration tests.

    ### Production Usage

    You can use the memory backend in production if you want, although there are some important caveats to keep
    in mind:

     1. There is limited attempts at performance optimization, so you should test it before putting it under
        substantial loads.
     2. There's no concept of replication.  If you have multiple workers, then write operations won't be
        persisted between them.

    So, while there may be some cases where this is useful in production, it's by no means a replacement for
    more typical in-memory data stores.

    ### Predefined Records

    You can declare a binding named `memory_backend_default_data` to seed the memory backend with records.  This
    can be helpful in testing to setup your tests, and is occassionally helpful for keeping track of data in
    fixed, read-only tables.  Here's an example:

    ```python
    import clearskies


    class Owner(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        phone = clearskies.columns.Phone()


    class Pet(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        species = clearskies.columns.String()
        owner_id = clearskies.columns.BelongsToId(Owner, readable_parent_columns=["id", "name"])
        owner = clearskies.columns.BelongsToModel("owner_id")


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.List(
            model_class=Pet,
            readable_column_names=["id", "name", "species", "owner"],
            sortable_column_names=["id", "name", "species"],
            default_sort_column_name="name",
        ),
        classes=[Owner, Pet],
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": Owner,
                    "records": [
                        {"id": "1-2-3-4", "name": "John Doe", "phone": "555-123-4567"},
                        {"id": "5-6-7-8", "name": "Jane Doe", "phone": "555-321-0987"},
                    ],
                },
                {
                    "model_class": Pet,
                    "records": [
                        {"id": "a-b-c-d", "name": "Fido", "species": "Dog", "owner_id": "1-2-3-4"},
                        {"id": "e-f-g-h", "name": "Spot", "species": "Dog", "owner_id": "1-2-3-4"},
                        {
                            "id": "i-j-k-l",
                            "name": "Puss in Boots",
                            "species": "Cat",
                            "owner_id": "5-6-7-8",
                        },
                    ],
                },
            ],
        },
    )

    if __name__ == "__main__":
        wsgi()
    ```

    And if you invoke it:

    ```bash
    $ curl 'http://localhost:8080' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "a-b-c-d",
                "name": "Fido",
                "species": "Dog",
                "owner": {
                    "id": "1-2-3-4",
                    "name": "John Doe"
                }
            },
            {
                "id": "i-j-k-l",
                "name": "Puss in Boots",
                "species": "Cat",
                "owner": {
                    "id": "5-6-7-8",
                    "name": "Jane Doe"
                }
            },
            {
                "id": "e-f-g-h",
                "name": "Spot",
                "species": "Dog",
                "owner": {
                    "id": "1-2-3-4",
                    "name": "John Doe"
                }
            }
        ],
        "pagination": {
            "number_results": 3,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }
    ```
    """

    default_data = inject.ByName("memory_backend_default_data")
    default_data_loaded = False
    _tables: dict[str, MemoryTable] = {}
    _silent_on_missing_tables: bool = True

    def __init__(self, silent_on_missing_tables=True):
        self.__class__._tables = {}
        self._silent_on_missing_tables = silent_on_missing_tables

    @classmethod
    def clear_table_cache(cls):
        cls._tables = {}

    def load_default_data(self):
        if self.default_data_loaded:
            return
        self.default_data_loaded = True
        if not isinstance(self.default_data, list):
            raise TypeError(
                f"'memory_backend_default_data' should be populated with a list, but I received a value of type '{self.default_data.__class__.__name__}'"
            )
        for index, table_data in enumerate(self.default_data):
            if "model_class" not in table_data:
                raise TypeError(
                    f"Each entry in the 'memory_backend_default_data' list should have a key named 'model_class', but entry #{index + 1} is missing this key."
                )
            model_class = table_data["model_class"]
            if not functional.validations.is_model_class(table_data["model_class"]):
                raise TypeError(
                    f"The 'model_class' key in 'memory_backend_default_data' for entry #{index + 1} is not a model class."
                )
            if "records" not in table_data:
                raise TypeError(
                    f"Each entry in the 'memory_backend_default_data' list should have a key named 'records', but entry #{index + 1} is missing this key."
                )
            records = table_data["records"]
            if not isinstance(records, list):
                raise TypeError(
                    f"The 'records' key in 'memory_backend_default_data' for entry #{index + 1} was not a list."
                )
            self.create_table(model_class)
            for record in records:
                self.create_with_model_class(record, model_class)

    def silent_on_missing_tables(self, silent=True):
        self._silent_on_missing_tables = silent

    def create_table(self, model_class: type[clearskies.model.Model]):
        self.load_default_data()
        table_name = model_class.destination_name()
        if table_name in self.__class__._tables:
            return
        self.__class__._tables[table_name] = MemoryTable(model_class)

    def has_table(self, model_class: type[clearskies.model.Model]) -> bool:
        self.load_default_data()
        table_name = model_class.destination_name()
        return table_name in self.__class__._tables

    def get_table(self, model_class: type[clearskies.model.Model], create_if_missing=False) -> MemoryTable:
        table_name = model_class.destination_name()
        if table_name not in self.__class__._tables:
            if create_if_missing:
                self.create_table(model_class)
            else:
                raise ValueError(
                    f"The memory backend was asked to work with the model '{model_class.__name__}' but this model hasn't been explicitly added to the memory backend.  This typically means that you are querying for records in a model but haven't created any yet."
                )
        return self.__class__._tables[table_name]

    def create_with_model_class(
        self, data: dict[str, Any], model_class: type[clearskies.model.Model]
    ) -> dict[str, Any]:
        self.create_table(model_class)
        return self.get_table(model_class).create(data)

    def update(self, id: int | str, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        self.create_table(model.__class__)
        return self.get_table(model.__class__).update(id, data)

    def create(self, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        self.create_table(model.__class__)
        return self.get_table(model.__class__).create(data)

    def delete(self, id: int | str, model: clearskies.model.Model) -> bool:
        self.create_table(model.__class__)
        return self.get_table(model.__class__).delete(id)

    def count(self, query: clearskies.query.Query) -> int:
        self.check_query(query)
        if not self.has_table(query.model_class):
            if self._silent_on_missing_tables:
                return 0

            raise ValueError(
                f"Attempt to count records for model '{query.model_class.__name__}' that hasn't yet been loaded into the MemoryBackend"
            )

        # this is easy if we have no joins, so just return early so I don't have to think about it
        if not query.joins:
            return self.get_table(query.model_class).count(query)

        # we can ignore left joins when counting
        query.joins = [join for join in query.joins if join.join_type != "LEFT"]
        return len(self.rows_with_joins(query))

    def records(
        self, query: clearskies.query.Query, next_page_data: dict[str, str | int] | None = None
    ) -> list[dict[str, Any]]:
        self.check_query(query)
        if not self.has_table(query.model_class):
            if self._silent_on_missing_tables:
                return []

            raise ValueError(
                f"Attempt to fetch records for model '{query.model_class.__name__} that hasn't yet been loaded into the MemoryBackend"
            )

        # this is easy if we have no joins, so just return early so I don't have to think about it
        if not query.joins:
            return self.get_table(query.model_class).rows(query, query.conditions, next_page_data=next_page_data)
        rows = self.rows_with_joins(query)

        if query.sorts:
            default_table_name = query.model_class.destination_name()
            rows = sorted(
                rows, key=cmp_to_key(lambda row_a, row_b: _sort(row_a, row_b, query.sorts, default_table_name))
            )

        # currently we don't do much with selects, so just limit results down to the data from the original
        # table.
        rows = [row[query.model_class.destination_name()] for row in rows]

        if "start" in query.pagination or query.limit:
            number_rows = len(rows)
            start = query.pagination.get("start", 0)
            if start >= number_rows:
                start = number_rows - 1
            end = len(rows)
            if query.limit and start + query.limit <= number_rows:
                end = start + query.limit
            rows = rows[start:end]
            if end < number_rows and type(next_page_data) == dict:
                next_page_data["start"] = start + query.limit
        return rows

    def rows_with_joins(self, query: clearskies.query.Query) -> list[dict[str, Any]]:
        joins = [*query.joins]
        conditions = [*query.conditions]
        # quick sanity check
        for join in query.joins:
            if join.unaliased_table_name not in self.__class__._tables:
                if not self._silent_on_missing_tables:
                    raise ValueError(
                        f"Join '{join._raw_join}' refrences table '{join.unaliased_table_name}' which does not exist in MemoryBackend"
                    )
                return []

        # start with the matches in the main table
        left_table_name = query.model_class.destination_name()
        left_conditions = self.conditions_for_table(left_table_name, conditions, is_left=True)
        main_rows = self.get_table(query.model_class).rows(query, left_conditions, filter_only=True)
        # and now adjust the way data is stored in our rows list to support the joining process.
        # we're going to go from something like: `[{row_1}, {row_2}]` to something like:
        # [{table_1: table_1_row_1, table_2: table_2_row_1}, {table_1: table_1_row_2, table_2: table_2_row_2}]
        # etc...
        rows = [{left_table_name: row} for row in main_rows]
        joined_tables = [left_table_name]

        # and now work through our joins.  The tricky part is order - we need to manage the joins in the
        # proper order, but they may not be in the correcet order in our join list.  I still don't feel like building
        # a full graph, so cheat and be dumb: loop through them all and join in the ones we can, skipping the ones
        # we can't.  If we get to the end and there are still joins left in the queue, then repeat, and eventually
        # complain (since the joins may not be a valid object graph)
        for i in range(10):
            for index, join in enumerate(joins):
                left_table_name = join.left_table_name
                alias = join.alias
                right_table_name = join.right_table_name
                table_name_for_join = alias if alias else right_table_name
                if left_table_name not in joined_tables:
                    continue

                join_rows = self.__class__._tables[join.unaliased_table_name].rows(query, [], filter_only=True)

                rows = self.join_rows(rows, join_rows, join, joined_tables)

                # done with this one!
                del joins[index]
                joined_tables.append(table_name_for_join)

            # are we done yet?
            if not joins:
                break

        if joins:
            raise ValueError(
                "Unable to fulfill joins for query - perhaps a necessary join is missing? "
                + "One way to get this error is if you tried to join on another table which hasn't been "
                + "joined itself.  e.g.: SELECT * FROM users JOIN type ON type.id=categories.type_id"
            )

        # now apply any remaining conditions.
        left_condition_ids = [id(condition) for condition in left_conditions]
        for condition in [condition for condition in conditions if id(condition) not in left_condition_ids]:
            condition_filter = MemoryTable._condition_as_filter(condition)
            rows = list(
                filter(lambda row: condition.table_name in row and condition_filter(row[condition.table_name]), rows)
            )

        return rows

    def all_rows(self, table_name: str) -> list[dict[str, Any]]:
        if table_name not in self.__class__._tables:
            if self._silent_on_missing_tables:
                return []

            raise ValueError(f"Cannot return rows for unknown table '{table_name}'")
        return list(filter(None, self.__class__._tables[table_name]._rows))

    def check_query(self, query: clearskies.query.Query) -> None:
        if query.group_by:
            raise KeyError(
                f"MemoryBackend does not support group_by clauses in queries. You may be using the wrong backend."
            )

    def conditions_for_table(
        self, table_name: str, conditions: list[clearskies.query.Condition], is_left=False
    ) -> list[clearskies.query.Condition]:
        """
        Return only the conditions for the given table.

        If you set is_left=True then it assumes this is the "default" table and so will also return conditions
        without a table name.
        """
        return [
            condition
            for condition in conditions
            if condition.table_name == table_name or (is_left and not condition.table_name)
        ]

    def join_rows(
        self,
        rows: list[dict[str, Any]],
        join_rows: list[dict[str, Any]],
        join: clearskies.query.Join,
        joined_tables: list[str],
    ) -> list[dict[str, Any]]:
        """
        Add the rows in `join_rows` in to the `rows` holder.

        `rows` should be something like:

        ```python
        [
            {
                "table_1": {"table_1_row_1"},
                "table_2": {"table_2_row_1"},
            },
            {
                "table_1": {"table_1_row_2"},
                "table_2": {"table_2_row_2"},
            },
        ]
        ```

        and join_rows should be the rows for the new table, something like:

        `[{table_3_row_1}, {table_3_row_2}]`

        which will then get merged into the rows variable properly (which it will return as a new list)
        """
        join_table_name = join.alias if join.alias else join.right_table_name
        join_type = join.join_type

        #######
        ########
        ## our problem is here.  When we join rows we can end up with multiple copies of the records from the left table because
        # there can be more than one matching record in the right table.  This isn't happening, and so we're not getting the
        # proper results because the one record that is chosen to match with the left table doesn't meet the where condition
        # that is applied at the very end.  If we have multiple records that match, they all need to get retunred in the
        # final list of rows here, so we can properly search everything.

        # loop through each entry in rows, find a matching table in join_rows, and take action depending on join type
        rows = [*rows]
        matched_right_row_indexes = set()
        left_table_name = join.left_table_name
        left_column_name = join.left_column_name
        # we're
        for row_index in range(len(rows)):
            row = rows[row_index]
            matching_row = None
            if left_table_name not in row:
                raise ValueError("Attempted to check join data from unjoined table, which should not happen...")
            left_value = (
                row[left_table_name][left_column_name]
                if (row[left_table_name] is not None and left_column_name in row[left_table_name])
                else None
            )
            matching_rows = []
            for join_index, join_row in enumerate(join_rows):
                right_value = join_row[join.right_column_name] if join.right_column_name in join_row else None
                # for now we are assuming the operator for the matching is `=`.  This is mainly because
                # our join parsing doesn't bother checking for the matching operator, because it is `=` in
                # 99% of cases.  We can always adjust down the line.
                if (right_value is None and left_value is None) or (right_value == left_value):
                    matching_rows.append(join_row)
                    matched_right_row_indexes.add(right_value)

            # if we have matching rows then join them in.
            for index, matching_row in enumerate(matching_rows):
                if not index:
                    rows[row_index][join_table_name] = matching_row
                else:
                    rows.append({**row, **{join_table_name: matching_row}})

            # if we don't have matching rows then remove them for an inner or right join
            if not matching_rows:
                if join_type == "LEFT" or join_type == "OUTER":
                    rows[row_index][join_table_name] = matching_row = None
                else:
                    # we can't immediately delete the row because we're looping over the array it is in,
                    # so just mark it as None and remove it later
                    rows[row_index] = None  # type: ignore

        rows = [row for row in rows if row is not None]

        # now for outer/right rows we add on any unmatched rows
        if (join_type == "OUTER" or join_type == "RIGHT") and len(matched_right_row_indexes) < len(join_rows):
            for join_index in set(range(len(join_rows))) - matched_right_row_indexes:
                rows.append(
                    {
                        join_table_name: join_rows[join_index],
                        **{table_name: None for table_name in joined_tables},
                    }
                )

        return rows

    def validate_pagination_data(self, data: dict[str, Any], case_mapping: Callable) -> str:
        extra_keys = set(data.keys()) - set(self.allowed_pagination_keys())
        if len(extra_keys):
            key_name = case_mapping("start")
            return "Invalid pagination key(s): '" + "','".join(extra_keys) + f"'.  Only '{key_name}' is allowed"
        if "start" not in data:
            key_name = case_mapping("start")
            return f"You must specify '{key_name}' when setting pagination"
        start = data["start"]
        try:
            start = int(start)
        except:
            key_name = case_mapping("start")
            return f"Invalid pagination data: '{key_name}' must be a number"
        return ""

    def allowed_pagination_keys(self) -> list[str]:
        return ["start"]

    def documentation_pagination_next_page_response(self, case_mapping: Callable[[str], str]) -> list[Any]:
        return [AutoDocInteger(case_mapping("start"), example=0)]

    def documentation_pagination_next_page_example(self, case_mapping: Callable[[str], str]) -> dict[str, Any]:
        return {case_mapping("start"): 0}

    def documentation_pagination_parameters(
        self, case_mapping: Callable[[str], str]
    ) -> list[tuple[AutoDocSchema, str]]:
        return [
            (
                AutoDocInteger(case_mapping("start"), example=0),
                "The zero-indexed record number to start listing results from",
            )
        ]



import os.path
from typing import Any


class Environment:
    """
    This loads up the environment configuration for the application.

    It looks in 3 possible places: first, it looks in os.environ.  Next, it tries to load up the .env file.
    Therefore, the application root directory should be passed in, at will look for a .env file there.
    It should contain lines like NAME=value.  Finally, if there is a value of `secret://path/to/secret`,
    it will use the secret service to look up the secret value.

    It is a very basic parser.  Empty lines and lines starting with a # will be ignored.  Otherwise everything
    is assumed to be a string.
    """

    _env_file_config: dict[str, Any] = None  # type: ignore
    _resolved_values: dict[str, Any] = {}
    os_environ: dict[str, Any] = {}

    def __init__(self, env_file_path, os_environ, secrets):
        self._env_file_path = env_file_path
        self.os_environ = os_environ
        self.secrets = secrets
        self._resolved_values = {}

    def get(self, name, silent=False) -> Any:
        self._load_env_file()
        if name in self.os_environ:
            return self.resolve_value(self.os_environ[name])
        if name in self._env_file_config:
            return self.resolve_value(self._env_file_config[name])

        if not silent:
            raise KeyError(f"Could not find environment config '{name}' in environment or .env file")
        return None

    def _load_env_file(self):
        if self._env_file_config is not None:
            return

        self._env_file_config = {}
        if not os.path.isfile(self._env_file_path):
            return

        with open(self._env_file_path, "r") as env_file:
            line_number = 0
            for line in env_file.readlines():
                line_number += 1
                (key, value) = self._parse_env_line(line, line_number)
                if key is None:
                    continue

                self._env_file_config[key] = value

    def _parse_env_line(self, line, line_number):
        line = line.strip()
        if not line:
            return (None, None)
        if line[0] == "#":
            return (None, None)
        if not "=" in line:
            raise ValueError(f"Parse error in environment line #{line_number}: should be 'key=value'")

        equal_index = line.index("=")
        key = line[:equal_index].strip()
        value = line[equal_index + 1 :].strip()
        lc_value = value.lower()
        if lc_value == "true":
            return (key, True)
        if lc_value == "false":
            return (key, False)
        if lc_value[0] == '"' and lc_value[-1] == '"':
            return (key, value.strip('"'))
        if lc_value[0] == "'" and lc_value[-1] == "'":
            return (key, value.strip("'"))
        try:
            as_int = int(value)
            return (key, as_int)
        except:
            pass
        try:
            as_float = float(value)
            return (key, as_float)
        except:
            pass
        return (key, value)

    def resolve_value(self, value):
        if type(value) != str or value[:9] != "secret://":
            return value

        secret_path = value[9:]
        if secret_path[0] != "/":
            secret_path = f"/{secret_path}"
        if secret_path not in self._resolved_values:
            if not self.secrets:
                raise ValueError(
                    "References to the secret engine were found in the environment, "
                    + "but a secret engine was not provided"
                )
            self._resolved_values[secret_path] = self.secrets.get(secret_path)
        return self._resolved_values[secret_path]



from clearskies.security_headers.cache_control import CacheControl
from clearskies.security_headers.cors import Cors
from clearskies.security_headers.csp import Csp
from clearskies.security_headers.hsts import Hsts

__all__ = [
    "CacheControl",
    "Cors",
    "Csp",
    "Hsts",
]



class Authorization:
    def gate(self, authorization_data, input_output):
        """
        Return True/False to denote if the given user, as represented by the authorization data, should be allowed access.

        Raise clearskies.exceptions.ClientError if you want to raise a specific error message.
        """
        return True

    def filter_model(self, model, authorization_data, input_output):
        """Return a models object with additional filters applied to account for authorization needs."""
        return model



class Default:
    def __init__(self, schema):
        self.schema = schema

    def convert(self):
        schema = {"type": self.schema._type}
        if self.schema._format:
            schema["format"] = self.schema._format
        if hasattr(self.schema, "example") and self.schema.example:
            schema["example"] = self.schema.example
        if hasattr(self.schema, "value") and self.schema.value:
            schema["example"] = self.schema.value
        return schema



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Self, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Integer as AutoDocInteger
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.column import Column
from clearskies.query import Condition

if TYPE_CHECKING:
    from clearskies import Model


class Integer(Column):
    """
    A column that stores integer data.

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        age = clearskies.columns.Integer()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["age"],
            readable_column_names=["id", "age"],
        ),
        classes=[MyModel],
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080' -d '{"age":20}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "6ea74719-a65f-45ae-b6a3-641ce682ed25",
            "age": 20
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080' -d '{"age":"asdf"}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "age": "value should be an integer"
        }
    }
    ```
    """

    default = configs.Integer(default=None)  #  type: ignore
    setable = configs.IntegerOrCallable(default=None)  #  type: ignore
    _allowed_search_operators = ["<=>", "!=", "<=", ">=", ">", "<", "=", "in", "is not null", "is null"]

    auto_doc_class: type[AutoDocSchema] = AutoDocInteger

    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        default: int | None = None,
        setable: int | Callable[..., int] | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> int:
        pass

    def __get__(self, instance, cls):
        if instance is None:
            self.model_class = cls
            return self

        value = super().__get__(instance, cls)
        return None if value is None else int(value)

    def __set__(self, instance, value: int) -> None:
        instance._next_data[self.name] = value

    def from_backend(self, value) -> int | None:
        return None if value is None else int(value)

    def to_backend(self, data):
        if self.name not in data or data[self.name] is None:
            return data

        return {**data, self.name: int(data[self.name])}

    def input_error_for_value(self, value, operator=None):
        try:
            int(value)
        except ValueError:
            return "value should be an integer"
        return ""

    def equals(self, value: int) -> Condition:
        return super().equals(value)

    def spaceship(self, value: int) -> Condition:
        return super().spaceship(value)

    def not_equals(self, value: int) -> Condition:
        return super().not_equals(value)

    def less_than_equals(self, value: int) -> Condition:
        return super().less_than_equals(value)

    def greater_than_equals(self, value: int) -> Condition:
        return super().greater_than_equals(value)

    def less_than(self, value: int) -> Condition:
        return super().less_than(value)

    def greater_than(self, value: int) -> Condition:
        return super().greater_than(value)

    def is_in(self, values: list[int]) -> Condition:
        return super().is_in(values)
